/** @file el5xxx_device.c Implementation of the device handler of the EL5xxx devices for the SOEM master. */

#include "stdincludes.h"
#include <limits.h>
#include "soem_devices.h"
#include "memory.h"

/* {{{ EL5002 */
#define EL5002_DEV_ID           0x138a3052
#define EL5002_DEV_STR          "EL5002"

typedef struct PACKED
{
    uint16_t status0;
    uint32_t channel0;
    uint16_t status1;
    uint32_t channel1;
} in_EL5002t;

static double el5002_read_enc0(IoInstance *io_instance) {
    DeviceInstance *soem_di = (DeviceInstance *)io_instance->device;
    const in_EL5002t *input_device = soem_di->input_address;

    return input_device->channel0;
}

static double el5002_read_enc1(IoInstance *io_instance) {
    DeviceInstance *soem_di = (DeviceInstance *)io_instance->device;
    const in_EL5002t *input_device = soem_di->input_address;

    return input_device->channel1;
}

static InputIoDefinition el5002_enc0 = {
    {DK_ENC, "enc0", 1}, &el5002_read_enc0
};
static InputIoDefinition el5002_enc1 = {
    {DK_ENC, "enc1", 1}, &el5002_read_enc1
};

static const InputIoDefinition *el5002_inputs[] = {&el5002_enc0, &el5002_enc1, NULL};
static const OutputIoDefinition *el5002_outputs[] = {NULL};
const DeviceDefinition el5002_device = {
    EL5002_DEV_STR,
    el5002_inputs, el5002_outputs,
    sizeof(DeviceInstance),
    NULL,
    EL5002_DEV_ID,
};

/* }}} */
/* {{{ EL5101 */
#define EL5101_NCHAN            1
#define EL5101_DEV_ID           0x13ed3052
#define EL5101_DEV_STR          "EL5101"

typedef struct PACKED
{
    uint8_t control;
    int16_t outvalue;
} out_EL5101t;

typedef struct PACKED
{
    uint8_t status;
    int16_t invalue;
    int16_t latch;
} in_EL5101t;

struct EL5101_DeviceInstanceStruct {
    DeviceInstance device_instance;

    int32_t cumulative_position;
    int16_t last_measurement;
};
typedef struct EL5101_DeviceInstanceStruct EL5101_DeviceInstance;

static void el5101_setup(DeviceInstance *soem_di) {
    EL5101_DeviceInstance *instance = (EL5101_DeviceInstance *)soem_di;

    instance->cumulative_position = INT32_MAX;
}

static double el5101_read_enc0(IoInstance *io_instance) {
    EL5101_DeviceInstance *el5101_di = (EL5101_DeviceInstance *)io_instance->device;
    const in_EL5101t *input_address = el5101_di->device_instance.input_address;

    /* First call, initialize variables. */
    if (el5101_di->cumulative_position == INT32_MAX) {
        el5101_di->cumulative_position = 0;
        el5101_di->last_measurement = input_address->invalue;
    }

    /* Assume real movement is less than 32k. */
    int16_t movement = input_address->invalue - el5101_di->last_measurement;
    el5101_di->cumulative_position += movement;
    el5101_di->last_measurement = input_address->invalue;

    return el5101_di->cumulative_position;
}

static const InputIoDefinition el5101_enc0 = {
    {DK_ENC, "enc0", 1}, &el5101_read_enc0
};

static const InputIoDefinition *el5101_inputs[] = {&el5101_enc0, NULL};
static const OutputIoDefinition *el5101_outputs[] = {NULL};

const DeviceDefinition el5101_device = {
    EL5101_DEV_STR,
    el5101_inputs, el5101_outputs,
    sizeof(EL5101_DeviceInstance),
    &el5101_setup,
    EL5101_DEV_ID,
};
/* }}} */
/* {{{ EL5152 */
#define EL5152_NCHAN            2
#define EL5152_DEV_ID           0x14203052
#define EL5152_DEV_STR          "EL5152"

typedef struct PACKED
{
    uint16_t control0;
    uint32_t counterval0;
    uint16_t control1;
    uint32_t counterval1;
} out_EL5152t;

typedef struct PACKED
{
    uint16_t status0;
    uint32_t channel0;
    uint32_t period0;
    uint16_t status1;
    uint32_t channel1;
    uint32_t period1;
} in_EL5152t;

/** Data of the EL5152 device */
struct EL5152_DeviceInstanceStruct {
    DeviceInstance soem_device_instance; /**< Standard SOEM device information. */
    int64_t position0; /**< Base position of the first encoder, or \c INT64_MAX if not initialized. */
    int64_t position1; /**< Base position of the second encoder, or \c INT64_MAX if not initialized. */
};
typedef struct EL5152_DeviceInstanceStruct EL5152_DeviceInstance;

static void el5152_setup(DeviceInstance *device_instance) {
    EL5152_DeviceInstance *instance = (EL5152_DeviceInstance *)device_instance;
    instance->position0 = INT64_MAX; /* Value cannot be obtained by reading a 32 bit value. */
    instance->position1 = INT64_MAX;
}

/**
 * Initialize both encoder positions from the first measurement.
 * @param di EL5152 device instance for this device.
 * @param input_address Data block read by the master.
 */
static void el5152_firstread(EL5152_DeviceInstance *di, const in_EL5152t *input_address) {
    di->position0 = input_address->channel0;
    di->position1 = input_address->channel1;
}

/**
 * Read first encoder of the EL5152.
 * @param io_instance Io instance of the first encoder.
 * @return The read value, a 32 bit signed offset relative to the first measurement.
 */
static double el5152_read_enc0(IoInstance *io_instance) {
    EL5152_DeviceInstance *el5152_di = (EL5152_DeviceInstance *)io_instance->device;
    const in_EL5152t *input_address = el5152_di->soem_device_instance.input_address;

    if (el5152_di->position0 == INT64_MAX) el5152_firstread(el5152_di, input_address);
    return input_address->channel0 - el5152_di->position0;
}

/**
 * Read second encoder of the EL5152.
 * @param io_instance Io instance of the second encoder.
 * @return The read value, a 32 bit signed offset relative to the first measurement.
 */
static double el5152_read_enc1(IoInstance *io_instance) {
    EL5152_DeviceInstance *el5152_di = (EL5152_DeviceInstance *)io_instance->device;
    const in_EL5152t *input_address = el5152_di->soem_device_instance.input_address;

    if (el5152_di->position1 == INT64_MAX) el5152_firstread(el5152_di, input_address);
    return input_address->channel1 - el5152_di->position1;
}

static const InputIoDefinition el5152_enc0 = {
    {DK_ENC, "enc0", 1}, &el5152_read_enc0
};

static const InputIoDefinition el5152_enc1 = {
    {DK_ENC, "enc1", 1}, &el5152_read_enc1
};

static const InputIoDefinition *el5152_inputs[] = {&el5152_enc0, &el5152_enc1, NULL};
static const OutputIoDefinition *el5152_outputs[] = {NULL};
const DeviceDefinition el5152_device = {
    EL5152_DEV_STR,
    el5152_inputs, el5152_outputs,
    sizeof(EL5152_DeviceInstance),
    el5152_setup,
    EL5152_DEV_ID,
};
/* }}} */
