/** @file error_storage.c Error storage and retrieval code. */

#include "stdincludes.h"
#include "error_storage.h"

/**
 * Initialize the error storage.
 * @param es Storage to initialize.
 */
void InitErrorStorage(ErrorStorage *es)
{
    es->message = NULL;
    es->group_index = -1;
    es->error = FALSE;
}

/**
 * Store a warning in the message if possible. Routine does not overwrite an already existing message.
 * @param es Destination of the warning message.
 * @param message Statically allocated message to store.
 * @param group_index Group number where the error is found (if non-negative).
 */
void SetWarning(ErrorStorage *es, const char *message, int16_t group_index)
{
    if (HasMessage(es)) return; /* Already filled with an error message, or a previous warning message. */

    es->message = message;
    es->group_index = group_index;
    es->error = FALSE;
}

/**
 * Store an error in the message if possible. Routine does not overwrite an already existing message.
 * @param es Destination of the error message.
 * @param message Statically allocated message to store.
 * @param group_index Group number where the error is found (if non-negative).
 */
void SetError(ErrorStorage *es, const char *message, int16_t group_index)
{
    if (HasErrorMessage(es)) return; /* Don't overwrite an earlier message. */

    es->message = message;
    es->group_index = group_index;
    es->error = TRUE;
}

/**
 * Copy a stored message.
 * @param dst Destination to write to.
 * @param src Origin to copy from.
 */
void CopyErrorStorage(ErrorStorage *dst, const ErrorStorage *src)
{
    dst->message = src->message;
    dst->group_index = src->group_index;
    dst->error = src->error;
}

