/** @file filterblock.c Implementation of the filter. */

#include "stdincludes.h"
#include "filterblock.h"
#include "memory.h"

/**
 * Initialize the filter. The filter needs weights to function properly.
 * @param f Filter to initialize.
 * @return Result of the operation.
 */
void FilterBlockInit(FilterBlock *f) {
    f->weights = NULL;
    f->samples = NULL;
    f->num_weights = 0;
    f->sample_offset = 0;
    f->operational = FALSE;
}

/**
 * Set the weights of the filter.
 * @param f Filter to set up.
 * @param mem Memory manager.
 * @param weights Array of \a num_weights weights.
 * @param num_weights Number of weights in the \a weights array.
 * @return Result of the operation.
 * @note First weight is applied to newest sample, second weight to second newest sample, and so on.
 */
BOOL FilterBlockSetWeights(FilterBlock *f, Memory *mem, const double *weights, size_t num_weights)
{
    if (num_weights <= 0 || weights == NULL) return FALSE;

    f->weights = MemoryAllocate(mem, num_weights * 2 * sizeof(double));
    if (f->weights == NULL) return FALSE;

    f->samples = &f->weights[num_weights];
    f->num_weights = num_weights;
    f->sample_offset = 0;
    f->operational = 0;

    /* Copy weights reversed as it simplifies computing the output, and initialize sanples to 0. */
    size_t i, j;
    j = num_weights - 1;
    for (i = 0; i < num_weights; i++) {
        f->weights[j] = weights[i];
        f->samples[i] = 0;
        j--;
    }
    return TRUE;
}

/**
 * Prepare filter for usage.
 * @param f Filter to prepare.
 * @return Result of the operation.
 */
BOOL FilterBlockPrecheck(FilterBlock *f)
{
    return f->num_weights > 0;
}

/**
 * Add a new sample to the filter.
 * @param f Filter to update.
 * @param sample New sample to add.
 * @return Whether the filter is operational (that is, a valid output can be computed from it).
 * @note Output gives valid results if \c b->operational holds after the call.
 */
BOOL FilterBlockAddSample(FilterBlock *f, double sample) {
    if(f->sample_offset >= f->num_weights) return FALSE; /* Ignore calls if no weights have been set. */

    f->samples[f->sample_offset] = sample;
    if (f->sample_offset + 1 < f->num_weights) {
        f->sample_offset = f->sample_offset + 1;
    } else {
        f->sample_offset = 0;
        f->operational = TRUE;
    }
    return f->operational;
}

/**
 * Compute the output of the filter.
 * @param f Filter to use.
 * @return Resulting output. Is valid only if \c f->operational holds.
 */
double FilterBlockComputeOutput(FilterBlock *f) {
    size_t weight_index = 0;
    size_t sample_index = f->sample_offset;

    double result = 0;
    /* sample_index >= weight_index, therefore it will hit the upper boundary sooner or equal. */
    while (sample_index < f->num_weights) {
        result += f->weights[weight_index] * f->samples[sample_index];
        weight_index++;
        sample_index++;
    }
    sample_index = 0;
    while (weight_index < f->num_weights) { /* Wrap sample_index around, and finish. */
        result += f->weights[weight_index] * f->samples[sample_index];
        weight_index++;
        sample_index++;
    }
    return result;
}

/**
 * Add a new sample to the filter, and compute the new output.
 * @param f Filter to update.
 * @param sample New sample to add.
 * @return Resulting output. Is valid only if \c f->operational holds after the call.
 */
double FilterBlockAddSampleCompute(FilterBlock *f, double sample) {
    FilterBlockAddSample(f, sample);
    return FilterBlockComputeOutput(f);
}

