/** @file function_generators.c Code for function generators. */

#include "stdincludes.h"
#include "function_generators.h"

/**
 * Initialize a sine function structure.
 * @param sf Sine function to initialize.
 * @param samples Sample array to use.
 * @param base Base value of the output.
 * @param amplitude Magnitude of the output.
 * @param num_samples Number of samples in \a samples.
 * @param index Start index in the sample array (normally 0).
 */
void InitSineFunction(SineFunction *sf, double *samples, double base, double amplitude,
                      uint32_t num_samples, uint32_t index)
{
    sf->samples = samples;
    sf->base = base;
    sf->amplitude = amplitude;
    sf->num_samples = num_samples;
    sf->index = index;
}

/**
 * Compute the next sample to produce for the given sine function, assuming a delay
 * has been performed of 'sample interval' length.
 * @param sf Sine function to use for computing the next value.
 * @return The next sample.
 * @note The \a sf is updated in-place.
 */
double GetSineFunctionSample(SineFunction *sf)
{
    double value = sf->base + sf->amplitude * sf->samples[sf->index];
    sf->index++;
    if (sf->index >= sf->num_samples) sf->index = 0;

    return value;
}

/**
 * Initialize a block function.
 * @param f Block function to initialize.
 * @param hi_length Length of the hi phase, in nsecs.
 * @param tot_length Length of the hi+lo phase, in nsecs.
 * @param hi_val Output value during the hi phase.
 * @param lo_val Output value during the lo phase.
 * @param offset Starting point in the block function (normally 0).
 */
void InitBlockFunction(BlockFunction *f, size_t hi_length, size_t tot_length,
                       double hi_val, double lo_val, size_t offset)
{
    f->hi_length = hi_length;
    f->total_length = tot_length;
    f->hi_value = hi_val;
    f->lo_value = lo_val;
    f->offset = offset;
}

/**
 * Compute the next sample to produce for the given block function, after performing
 * a delay of \a delay_step nsecs.
 * @param bf Block function to use for computing the next value.
 * @param delay_step Delay step performed just before the query.
 * @return The next sample.
 * @note The \a bf is updated in-place.
 */
double GetBlockFunctionSample(BlockFunction *bf, size_t delay_step)
{
    size_t offset = bf->offset + delay_step;
    while (offset >= bf->total_length) offset -= bf->total_length;

    bf->offset = offset;
    return (offset <= bf->hi_length) ? bf->hi_value : bf->lo_value;
}

/**
 * Initialize a triangle function.
 * @param f Triangle function to initialize.
 * @param up_len Length of the 'up' phase in nsecs.
 * @param down_len Length of the down phase, in nsecs.
 * @param high Highest value of the function.
 * @param low Lowest value of the function.
 * @param offset Starting point in the function (normally 0).
 */
void InitTriangleFunction(TriangleFunction *f, size_t up_len, size_t down_len,
                          double high, double low, size_t offset)
{
    f->up_length = up_len;
    f->down_length = down_len;
    f->hi_value = high;
    f->lo_value = low;
    f->offset = offset;
}

/**
 * Compute the next sample to produce for the given triangle function, after performing
 * a delay of \a delay_step nsecs.
 * @param tf Triangle function to use for computing the next value.
 * @param delay_step Delay step performed just before the query.
 * @return The next sample.
 * @note The \a tf is updated in-place.
 */
double GetTriangleFunctionSample(TriangleFunction *tf, size_t delay_step)
{
    size_t offset = tf->offset + delay_step;
    size_t total_length = tf->up_length + tf->down_length;
    while (offset >= total_length) offset -= total_length;

    tf->offset = offset;
    if (offset <= tf->up_length) {
        return tf->lo_value + (tf->hi_value - tf->lo_value) * offset / tf->up_length;
    } else {
        offset -= tf->up_length;
        return tf->hi_value + (tf->lo_value - tf->hi_value) * offset / tf->down_length;
    }
}

