/** @file function_generators.h Declarations for function generators. */

#ifndef FUNCTION_GENERATORS_H
#define FUNCTION_GENERATORS_H

struct SineFunctionStruct {
    double *samples; /**< (constant) Unscaled sine samples. */
    double base; /**< (constant) Base value of the sine. */
    double amplitude; /**< (constant) Amplitude of the sine. */
    uint32_t num_samples; /**< (constant) Number of samples in #samples. */
    uint32_t index; /**< Next sample in #samples to use. */
};
typedef struct SineFunctionStruct SineFunction;

/**
 * Function alternating between hi and low value.
 * If #offset is less than #hi_length, function is in hi phase. If #offset is
 * less than #total_length, it is in lo phase.
 */
struct BlockFunctionStruct {
    size_t hi_length; /**< (constant) Length of the hi phase, in nsecs. */
    size_t total_length; /**< (constant) Length of the hi+lo phase, in nsecs. */
    double hi_value; /**< (constant) Output value in the hi phase. */
    double lo_value; /**< (constant) Output value in the lo phase. */
    size_t offset; /**< Offset of the last output in hi or lo phase. */
};
typedef struct BlockFunctionStruct BlockFunction;

struct TriangleFunctionStruct {
    size_t up_length; /**< (constant) Length of the up phase, in nsecs. */
    size_t down_length; /**< (constant) Length of the down phase, in nsecs. */
    double hi_value; /**< (constant) Output value at the end of up phase. */
    double lo_value; /**< (constant) Output value at the end of down phase. */
    size_t offset; /**< Offset of the last output in up or down phase. */
};
typedef struct TriangleFunctionStruct TriangleFunction;

void InitBlockFunction(BlockFunction *f, size_t hi_length, size_t tot_length,
                       double hi_val, double lo_val, size_t offset);
void InitTriangleFunction(TriangleFunction *f, size_t up_len, size_t down_len,
                          double high, double low, size_t offset);
void InitSineFunction(SineFunction *sf, double *samples, double base, double amplitude,
                      uint32_t num_samples, uint32_t index);

double GetSineFunctionSample(SineFunction *sf);
double GetBlockFunctionSample(BlockFunction *bf, size_t delay_step);
double GetTriangleFunctionSample(TriangleFunction *tf, size_t delay_step);

#endif
