/** @file memory.c Fully managed memory manager implementation. */

#include "stdincludes.h"
#include "memory.h"

/**
 * Get the memory alignment requirement of any type.
 * @return Allocated memory should start at a multiple of the returned alignment.
 */
static size_t MemoryAlignment(void)
{
    /* Alignment of random structure definition should cover all. */
    size_t alignment = TYPE_ALIGNMENT(Memory);

    assert(alignment >= TYPE_ALIGNMENT(uint8_t));
    assert(alignment >= TYPE_ALIGNMENT(uint16_t));
    assert(alignment >= TYPE_ALIGNMENT(uint32_t));
    assert(alignment >= TYPE_ALIGNMENT(uint64_t));
    assert(alignment >= TYPE_ALIGNMENT(float));
    assert(alignment >= TYPE_ALIGNMENT(double));
    return alignment;
}

/**
 * Initialize managed memory for adding memory blocks. After
 * #MemorySetMemory, blocks of memory can be requested.
 * @param mem Memory manager data.
 */
void MemoryInit(Memory *mem)
{
    pthread_mutex_init(&mem->lock, NULL);

    mem->free_memory = NULL;
    mem->remaining_size = 0;
    mem->base = NULL;
}

/**
 * Clear all resources from the memory manager.
 * @param mem Memory manager data to clear.
 * @return Base address of the memory provided to the manager.
 */
void *MemoryDestroy(Memory *mem)
{
    void *base = mem->base;

    pthread_mutex_destroy(&mem->lock);
    mem->free_memory = NULL;
    mem->remaining_size = 0;
    mem->base = NULL;
    return base;
}

/**
 * Give a block of memory.
 * @param mem Memory manager data.
 * @param size Size of the requested memory.
 * @return Base address of the memory, may be \c NULL is \c 0 bytes are
 *      requested or if there is not enough memory available.
 */
static void *Allocate(Memory *mem, size_t size)
{
    void *data;
    const size_t alignment = MemoryAlignment();
    size_t remainder;

    if (size == 0) return NULL;

    /* Extend size to a multiple of 'alignment'. */
    remainder = size % alignment;
    if (remainder != 0) size += alignment - remainder;

    /* Claim memory. */
    if (mem->remaining_size < size) {
        data = NULL;
    } else {
        mem->remaining_size -= size;
        data = mem->free_memory;
        mem->free_memory += size;
    }

    return data;
}

/**
 * Give memory to the memory manager.
 * @param mem Memory manager data.
 * @param base Base address of memory available for use.
 * @param size Size of the given memory.
 */
void MemorySetMemory(Memory *mem, void *base, size_t size)
{
    mem->free_memory = base;
    mem->remaining_size = size;
    mem->base = base;

    /* Align memory on alignment boundary. */
    const size_t alignment = MemoryAlignment();
    uintptr_t addr = (uintptr_t)base;
    uintptr_t remaining = addr % alignment;
    if (remaining != 0) {
        mem->free_memory += alignment - remaining;
        mem->remaining_size -= alignment - remaining;
    }
}

/**
 * Query memory from the manager.
 * @param mem Memory manager data.
 * @param size Required amount of memory.
 * @return Base address of the memory, may be \c NULL is \c 0 bytes are
 *      requested or if there is not enough memory available.
 */
void *MemoryAllocate(Memory *mem, size_t size)
{
    void *address;

    pthread_mutex_lock(&mem->lock);
    address = Allocate(mem, size);
    pthread_mutex_unlock(&mem->lock);
    return address;
}

/**
 * Query all memory from the manager.
 * @param mem Memory manager data.
 * @param max_size [out] Amount of available memory.
 * @return Start address of still available memory.
 * @note The memory is not claimed until #MemoryAllocateSetNeeded is called.
 *      In the mean time, no other memory claims should be performed.
 */
void *MemoryAllocateAll(Memory *mem, size_t *max_size)
{
    const size_t alignment = MemoryAlignment();
    void *address;

    pthread_mutex_lock(&mem->lock);
    if (mem->remaining_size <= alignment) {
        pthread_mutex_unlock(&mem->lock);

        *max_size = 0;
        return NULL;
    }
    *max_size = mem->remaining_size - alignment;
    address = mem->free_memory;
    pthread_mutex_unlock(&mem->lock);
    return address;
}

/**
 * Update memory manager with the real required amount of memory.
 * @param mem Memory manager data.
 * @param start Start address previously given out.
 * @param size Actual required amount of memory, should be less or equal to previously stated limit.
 */
void MemoryAllocateSetNeeded(Memory *mem, void *start, size_t size)
{
    void *real_address;

    if (start == NULL || size == 0) return;

    pthread_mutex_lock(&mem->lock);
    if (start != mem->free_memory) {
        assert(FALSE); /* Mutex still locked, but the game is over anyway. */
        exit(1);
    }

    real_address = Allocate(mem, size);
    assert(real_address == start);
    pthread_mutex_unlock(&mem->lock);
}

