/** @file soem_devices.h Device definitions, instances, and io points under soem. */

#ifndef SOEM_DEVICES_H
#define SOEM_DEVICES_H

typedef struct DeviceInstanceStruct DeviceInstance;
typedef struct IoInstanceStruct IoInstance;
typedef struct MemoryStruct Memory;

/** Available kinds of input/output. */
enum IoKindEnum {
    DK_DIN,  /**< Digital input. */
    DK_AIN,  /**< Analogue input. */
    DK_ENC,  /**< Encoder/counter input. */

    DK_DOUT, /**< Digital output. */
    DK_AOUT, /**< Analogue output. */
    DK_PWM,  /**< Pulse width modulator. */

    DK_NUM_IO_KINDS, /**< Number of available io kinds. */

    DK_NONE = DK_NUM_IO_KINDS, /**< No kind specified. */
};
typedef enum IoKindEnum IoKind;

/** Generic part of an io definition. */
struct IoDefinitionStruct {
    IoKind kind;        /**< Kind of io. */

    const char *name;   /**< Name of the io data point. */
    int num_channels;   /**< Number of output channels of the io point. */
};
typedef struct IoDefinitionStruct IoDefinition;

/** Definition of an input data point. */
struct InputIoDefinitionStruct {
    IoDefinition io_definition; /** Generic part of an io definition. */
    double (*read)(IoInstance *io_instance); /**< Read function of the io point. */
};
typedef struct InputIoDefinitionStruct InputIoDefinition;

/** Definition of an output data point. */
struct OutputIoDefinitionStruct {
    IoDefinition io_definition; /** Generic part of an io definition. */
    void (*write)(IoInstance *io_instance, double val); /**< Write function of the io point. */
};
typedef struct OutputIoDefinitionStruct OutputIoDefinition;

/** Device known to the library. */
struct DeviceDefinitionStruct {
    const char *name; /**< Name of the device type. */
    const InputIoDefinition **input_io_definitions;   /**< Input io point definitions of the device. */
    const OutputIoDefinition **output_io_definitions; /**< Output io point definitions of the device. */
    size_t malloc_size;
    /* XXX default init??!! setup doesn't have that access!! */
    void (*setup)(DeviceInstance *dev_instance);

    uint32_t device_id; /**< Id of the device type in EtherCaT. */
};
typedef struct DeviceDefinitionStruct DeviceDefinition;

/** Instance of a device after detecting at at the EtherCaT network. */
struct DeviceInstanceStruct {
    const DeviceDefinition *device_definition; /**< Definition of the found device. */
    uint16_t instance_number;  /**< Number of the found device within its type. */

    const void *input_address; /**< Base address of data read by the master for this device instance. */
    void *output_address;      /**< Base address of data to be written by the master for this device instance. */
};

/** io point of a device instance. */
struct IoInstanceStruct {
    DeviceInstance *device; /**< Found device that contains this io point. */
    const IoDefinition *io_definition; /**< Definition of the io point at the instance. */
};

DeviceInstance **InstantiateDevices(Memory *mem, size_t *device_count);
IoInstance *InstantiateIoPoints(DeviceInstance **dev_instances, size_t dev_count, Memory *mem, size_t *num_iis);

IoInstance *FindIoInstance(IoInstance *iis, size_t num_iis, const char *dev_name, size_t dev_index, const char *io_point_name);
IoInstance *FindIoKind(IoInstance *iis, size_t num_iis, IoKind kind, size_t kind_index);

/**
 * Perform read access on an input io instance.
 * @param ii Io instance to read from.
 * @return The read value.
 */
#define SOEM_READ(ii) (((InputIoDefinition *)((ii)->io_definition))->read)(ii)

/**
 * Perform write access on an io instance.
 * @param ii Io instance to write to.
 * @param val Value to write.
 */
#define SOEM_WRITE(ii, val) (((OutputIoDefinition *)((ii)->io_definition))->write)((ii), (val))

#endif
