/** @file triggerblock.c Trigger implementation. */

#include "stdincludes.h"
#include "triggerblock.h"

/**
 * Initialize trigger treshold point.
 * @param t The trigger to configure.
 * @param treshold Treshold value to trigger.
 * @param trigger_up Trigger on upgoing signal.
 * @param trigger_down Trigger on declining signal.
 */
void TriggerBlockSetPoint(TriggerBlock *t, double treshold, BOOL trigger_up, BOOL trigger_down)
{
    if (trigger_up) {
        t->trigger_type = trigger_down ? TT_REACHING : TT_RISING;
    } else {
        t->trigger_type = trigger_down ? TT_DECLINING : TT_NONE;
    }
    t->treshold = treshold;
    t->previous_sample = treshold;
}

/**
 * Initialize the length of the enabled interval.
 * @param t The trigger to configure.
 * @param copy_count Number of hardware samples with 'enabled' output after matching
 *      the treshold conditions. Should be 1 or larger.
 */
void TriggerBlockSetCount(TriggerBlock *t, ssize_t copy_count)
{
    if (copy_count <= 0) copy_count = 1;

    t->copy_count = copy_count;
    t->counter = 0;
}

/**
 * Initialize trigger to default (reach 0, 1 sample interval enabled).
 * @param t The trigger to configure.
 */
void TriggerBlockInit(TriggerBlock *t)
{
    TriggerBlockSetPoint(t, 0, TRUE, TRUE);
    TriggerBlockSetCount(t, 1);
}

/**
 * Save a new sample in the trigger, without testing for triggering.
 * @param t Trigger that gets the new sample.
 * @param sample Sample to add.
 */
static void TriggerBlockSaveSample(TriggerBlock *t, double sample)
{
    t->previous_sample = sample;
}

/**
 * Store the new sample in the trigger, and test for trigger firing.
 * @param t Trigger that gets the new sample.
 * @param sample Sample to add.
 * @return Whether the trigger fires.
 */
static BOOL TriggerBlockStoreTest(TriggerBlock *t, double sample)
{
    float previous = t->previous_sample;
    t->previous_sample = sample;
    switch (t->trigger_type) {
        case TT_NONE:
            return FALSE;

        case TT_RISING:
            return previous < t->treshold && sample >= t->treshold;

        case TT_DECLINING:
            return previous > t->treshold && sample <= t->treshold;

        case TT_REACHING:
            return (previous < t->treshold && sample >= t->treshold) ||
                   (previous > t->treshold && sample <= t->treshold);

        default: NOT_REACHED();
    }
}

/**
 * Add a new sample to the trigger. Trigger saves the new sample, and moves sampling forward.
 * @param t Trigger that gets the new sample.
 * @param sample Sample to add.
 */
void TriggerBlockAddSample(TriggerBlock *t, double sample)
{
    /* If copy process is running, do not test for trigger firing. */
    if (t->counter >= 1) { /* Trigger is enabled, just store the new sample without checking. */
        t->counter--;
        TriggerBlockSaveSample(t, sample);
        return;
    }

    /* Copy process is not running, add sample, test for trigger firing. */
    if (TriggerBlockStoreTest(t, sample)) t->counter = t->copy_count;
}

