import sys

# {{{ class AbsCodeBox:
class AbsCodeBox:
    """
    Code box with absolute indenting.
    """
    def __init__(self):
        self.lines = []

    def abs_add(self, abs_spaces, text):
        """
        Add a line with given absolute indentation and content.
        """
        self.lines.append((abs_spaces, text))

    def abs_extend(self, abs_spaces, texts):
        for text in texts:
            self.lines.append((abs_spaces, text))

    def abs_add_box(self, abs_spaces, box):
        for box_indent, box_text in box.lines:
            self.lines.append((abs_spaces + box_indent, box_text))

    def ensure_empty(self):
        """
        Ensure there is no line with text at the end.
        """
        if len(self.lines) == 0 or self.lines[-1][1] == "":
            return

        self.lines.append((0, ""))

    def trim(self):
        """
        Trim empty lines from both sides.
        """
        first, last = 0, len(self.lines) - 1
        while first <= last and self.lines[first][1] == "":
            first = first + 1
        while last > first and self.lines[last][1] == "":
            last = last - 1

        if first > 0 or last < len(self.lines) - 1:
            self.lines = self.lines[first : last + 1]

    def empty(self, trim_first=False):
        """
        Check whether the box contains any lines.
        """
        if trim_first:
            self.trim()

        return len(self.lines) == 0

    def stream(self):
        """
        Output code as a generator stream.
        """
        prefixes = {}

        for abs_spaces, text in self.lines:
            if text == "":
                yield ""
                continue

            abs_spaces = max(0, abs_spaces)
            prefix = prefixes.get(abs_spaces)
            if prefix is None:
                prefix = " " * abs_spaces
                prefixes[abs_spaces] = prefix

            yield prefix + text
# }}}
# {{{ class CodeBox(AbsCodeBox):
class CodeBox(AbsCodeBox):
    def __init__(self, indent_size = 4):
        AbsCodeBox.__init__(self)
        self.indent_size = indent_size
        self.num_indent = 0

    def append(self, text=None, pre=0, post=0):
        """
        Compatibility function.
        """
        self.add(text, pre, post)

    def add(self, text=None, pre=0, post=0):
        """
        Add a line of code to the output, possibly adapring indenting before or after the line.

        @param text: Line to append.
        @param pre: Amount of indenting to change before writing the line.
        @param post: Amount of indenting to change after writing the line.
        """
        if text is None or text == '':
            self.abs_add(0, "")
            return

        if pre != 0:
            self.indent(pre)

        self.abs_add(self.indent_size * self.num_indent, text)

        if post != 0:
            self.indent(post)

    def extend(self, texts):
        self.abs_extend(self.indent_size * self.num_indent, texts)

    def add_box(self, box):
        self.abs_add_box(self.indent_size * self.num_indent, box)

    def indent(self, count=1):
        self.num_indent = self.num_indent + count

    def dedent(self, count=1):
        self.indent(-count)
# }}}
# {{{ class CppLineCodeBox(AbsCodeBox):
class CppLineCodeBox(AbsCodeBox):
    """
    Absolute code box, with automagic #line pragmas added on request.

    @ivar line_preagmas: Whether #line pragmas should be generated.
    @ivar fname: Name of the origin filename.
    @ivar lineno: Number of the line in L{fname}.
    """
    def __init__(self, line_preagmas=False):
        AbsCodeBox.__init__(self)
        self.line_preagmas = line_preagmas
        self.fname = None
        self.lineno = None

    def set_fname_line(self, fname, lineno):
        if not self.line_preagmas:
            return # Disable '#line' pagramss

        if fname != self.fname:
            AbsCodeBox.abs_add(self, 0, '#line {} "{}"'.format(lineno, fname))
            self.fname = fname
            self.lineno = lineno
            return

        if lineno != self.lineno:
            AbsCodeBox.abs_add(self, 0, '#line {}'.format(lineno))
            self.lineno = lineno
            return

        # Else all is fine, don't output anything.

    def next_line(self, count):
        if self.lineno is not None:
            self.lineno = self.lineno + count

    def abs_add(self, abs_spaces, text):
        AbsCodeBox.abs_add(self, abs_spaces, text)
        self.next_line(1)

    def abs_add_box(self, abs_spaces, box):
        AbsCodeBox.abs_add_box(self, abs_spaces, box)
        self.next_line(count=len(box.lines))
# }}}

def make_code_literal(lines):
    """
    Convert multi-line string literal to a code box with lines.
    """
    # Split lines
    box = CodeBox()
    for line in lines.splitlines():
        box.abs_add(0, line.rstrip())

    box.trim()
    return box


def write_code(box, fname):
    """
    Write the code in the L{box} to the given file.

    @param box: Code to write.
    @type  box: L{CodeBox}

    @param fname: File to write to, C{"-"} means 'stdout'.
    @type  fname: C{str}
    """
    if fname == '-':
        handle = sys.stdout
    else:
        handle = open(fname, 'wt')

    for line in box.stream():
        handle.write(line)
        handle.write('\n')

    if fname != '-':
        handle.close()
