"""
Functions for compiling the generated c file.
"""
import os, re, subprocess
from acqgenlib import report, code

def die_missing_acqgen_files():
    """
    Report about missing source code, and die
    """
    msg = "Cannot find the acqgen runtime code, please set 'acqgen_dir' and/or 'acqgen_sources'"
    report.report_error(None, msg)
    # Program dies

class CompileInformation:
    """
    All data needed to perform compilation.
    """
    def __init__(self):
        self.matlab_base = None
        self.matlab_mex = None
        self.soem_base = None
        self.soem_lib = None
        self.soem_lflags = None
        self.soem_include = None
        self.compiler = None
        self.acqgen_dir = None
        self.acqgen_sources = None
        self.cflags = None
        self.lflags = None
        self.buffered = None
        self.verbose = False

        self.soem_lflags_list = None
        self.acqgen_sources_list = None

    # {{{ def get_matlab_base_path(self):
    def get_matlab_base_path(self):
        if self.matlab_base is not None:
            return self.matlab_base

        ML_BASE_PATH = "/usr/local/MATLAB"
        if not os.path.isdir(ML_BASE_PATH):
            msg = "Cannot find the {} directory, please set 'matlab_base'".format(ML_BASE_PATH)
            report.report_error(None, msg)
            # Program dies

        ML_RELEASE_PATTERN = re.compile("[rR]2[0-9][0-9][0-9]")
        release_name = None
        for entry in os.listdir(ML_BASE_PATH):
            m = ML_RELEASE_PATTERN.match(entry)
            if not m:
                continue

            base_path = os.path.join(ML_BASE_PATH, entry)
            mex_path = os.path.join(base_path, "bin/mex")
            if not os.path.isfile(mex_path):
                continue

            if release_name is None or entry[1:] > release_name:
                release_name = entry[1:]
                self.matlab_base = base_path

        if self.matlab_base is None:
            msg = "Cannot find a Matlab program with 'mex', please set 'matlab_base'"
            report.report_error(None, msg)

        return self.matlab_base
    # }}}
    # {{{ def get_matlab_mex_path(self):
    def get_matlab_mex_path(self):
        if self.matlab_mex is not None:
            return self.matlab_mex

        self.matlab_mex = os.path.join(self.get_matlab_base_path(), "bin/mex")
        if not os.path.exists(self.matlab_mex):
            msg = "Cannot find 'mex', please set 'matlab_mex'"
            report.report_error(None, msg)
        return self.matlab_mex
    # }}}
    # {{{ def get_soem_base(self):
    def get_soem_base(self):
        if self.soem_base is not None:
            return self.soem_base

        soem_base_dir = code.get_soem_installed_dir()
        if soem_base_dir is not None:
            self.soem_base = soem_base_dir
            return self.soem_base

        msg = "Cannot find an installed SOEM, please set 'soem_base'"
        report.report_error(None, msg)
        # Program dies
    # }}}
    # {{{ def get_soem_lib(self):
    def get_soem_lib(self):
        if self.soem_lib is not None:
            return self.soem_lib

        self.soem_lib = os.path.join(self.get_soem_base(), "lib/libsoem.a")
        if not os.path.isfile(self.soem_lib):
            msg = "Cannot find the 'libsoem.a' archive, please set 'soem_lib'"
            report.report_error(None, msg)
            # Program dies

        return self.soem_lib
    # }}}
    # {{{ def get_soem_include(self):
    def get_soem_include(self):
        if self.soem_include is not None:
            return self.soem_include

        base = self.get_soem_base()

        # First, try $BASE/include/soem
        inc = os.path.join(base, "include/soem")
        if os.path.isfile(os.path.join(inc, "ethercat.h")):
            self.soem_include = inc
            return self.soem_include

        # Second, try $BASE/include
        inc = os.path.join(base, "include")
        if os.path.isfile(os.path.join(inc, "ethercat.h")):
            self.soem_include = inc
            return self.soem_include

        msg = "Cannot find the SOEM include directory, please set 'soem_include'"
        report.report_error(None, msg)
        # Program dies
    # }}}
    # {{{ def get_soem_lflags_list(self):
    def get_soem_lflags_list(self):
        if self.soem_lflags_list is not None:
            return self.soem_lflags_list

        if self.soem_lflags is not None:
            self.soem_lflags_list = self.soem_lflags.split()
            return self.soem_lflags_list

        # Get library file, and split in directory prefix and archive file.
        libpath = self.get_soem_lib()
        libdir, libfile = os.path.split(libpath)

        # Strip 'lib' prefix and '.a/.so' suffix
        if libfile.startswith("lib"):
            libfile = libfile[3:]
        if libfile.endswith(".a"):
            libfile = libfile[:-2]
        elif libfile.endswith(".so"):
            libfile = libfile[:-3]

        # Construct the usual -L<dir> -l<file> flags.
        self.soem_lflags_list = []
        self.soem_lflags_list.append("-L{}".format(libdir))
        self.soem_lflags_list.append("-l{}".format(libfile))
        return self.soem_lflags_list
    # }}}
    # {{{ def get_compiler(self):
    def get_compiler(self):
        if self.compiler is not None:
            return self.compiler

        self.compiler = self.get_matlab_mex_path()
        return self.compiler
    # }}}
    # {{{ def get_acqgen_dir(self):
    def get_acqgen_dir(self):
        if self.acqgen_dir is not None:
            return self.acqgen_dir

        self.acqgen_dir = code.get_source_dir()
        if self.acqgen_dir is not None:
            return self.acqgen_dir

        die_missing_acqgen_files()
    # }}}
    # {{{ def get_acqgen_sources(self):
    def get_acqgen_sources(self):
        """
        Get the set source files to compile.
        """
        if self.acqgen_sources_list is not None:
            return self.acqgen_sources_list

        if self.acqgen_sources is not None:
            source_files = self.acqgen_sources.split()

        else:
            assert self.buffered is not None
            if self.buffered:
                source_files = code.get_buffered_code_files()
            else:
                source_files = code.get_unbuffered_code_files()


        source_dir = self.get_acqgen_dir()
        if source_dir is not None:
            self.acqgen_sources_list = [os.path.join(source_dir, sf) for sf in source_files]
        else:
            self.acqgen_sources_list = source_files
        return self.acqgen_sources_list
    # }}}
    # {{{ def get_cflags_list(self):
    def get_cflags_list(self):
        if self.cflags is None:
            src_dir = self.get_acqgen_dir()
            if src_dir is None:
                die_missing_acqgen_files()

            return ['-I{}'.format(src_dir), "-I{}".format(self.get_soem_include())]
        else:
            return self.cflags.split()
    # }}}
    # {{{ def get_lflags_list(self):
    def get_lflags_list(self):
        if self.lflags is None:
            return self.get_soem_lflags_list() + ["-lpthread", "-lm"]
        else:
            return self.lflags.split()
    # }}}

def build_compile_info(config):
    """
    Build the compile information object from the configuration block.
    """
    compile_info = CompileInformation()
    compile_info.matlab_base    = config.matlab_base
    compile_info.matlab_mex     = config.matlab_mex
    compile_info.soem_base      = config.soem_base
    compile_info.soem_lib       = config.soem_lib
    compile_info.soem_lflags    = config.soem_lflags
    compile_info.soem_include   = config.soem_include
    compile_info.compiler       = config.compiler
    compile_info.acqgen_dir     = config.acqgen_dir
    compile_info.acqgen_sources = config.acqgen_sources
    compile_info.cflags         = config.cflags
    compile_info.lflags         = config.lflags
    compile_info.buffered       = config.buffered
    return compile_info

def escape(word):
    """
    Make the L{word} mostly safe for parsing by the shell.
    """
    word = word.replace('\\', '\\\\')
    word = word.replace(' ', '\\ ')
    word = word.replace('&', '\\&')
    word = word.replace('"', '\\"')
    word = word.replace("'", "\\'")
    return word

def mex_cc(gen_filename, compile_info):
    """
    Run the 'mex' compilation.

    @param gen_filename: Name of the generated module C file.
    @param compile_info: Information where to find mex and dependencies such as
                         soem and runtime files and flags.
    """
    cmd = [compile_info.get_compiler()]
    cmd.extend(compile_info.get_cflags_list())
    cmd.append(gen_filename)
    cmd.extend(compile_info.get_acqgen_sources())
    cmd.extend(compile_info.get_lflags_list())

    if compile_info.verbose:
        # Dump command being executed onto the output, semi-safe for shell interpretation.
        last = len(cmd) - 1
        for i, word in enumerate(cmd):
            if i == last:
                word = escape(word)
            else:
                word = escape(word) + " \\"

            if i == 0:
                print(word)
            else:
                print("\t" + word)

    result = subprocess.run(cmd)
    if result.returncode != 0:
        msg = "C compilation failed!"
        report.report_error(None, msg)

