from acqgenlib import report, codebox, build_ast

def compute_sine_sampling(period, sample_interval, max_samples):
    """
    Compute best approximation for a sine function with frequency 1/L{period},
    given the sample interval at L{sample_interval}, with at most L{max_samples}
    pre-computed samples.

    @param period: Period of the sine function in seconds/cycle.
    @param sample_interval: Sample interval, in secs.
    @param max_samples: Maximum allowed number of samples.

    @return: Frequency of the approximated sine function, number of periods, number of samples.
    """
    def sampling_n(period, interval, n, max_samples):
        """
        Compute best solution to sample a sine with length L{period} for L{n}
        periods, with sample rate L{interval}.

        @return: Number of samples, remaining interval time (positive or negative).
        """
        total_period = period * n

        # Option 1, shorten the period a bit.
        floor_sample_count = int(total_period / interval)
        floor_remaining = total_period - interval * floor_sample_count

        # Option 2, extend the period a bit.
        ceil_sample_count = floor_sample_count + 1
        ceil_remaining = total_period - interval * ceil_sample_count

        if ceil_sample_count > max_samples:
            # Extended period is too many samples, the shortened version is the only solution.
            # Note that the shortened version may be too big as well.
            return floor_sample_count, floor_remaining

        if abs(floor_remaining) <= abs(ceil_remaining):
            return floor_sample_count, floor_remaining
        else:
            return ceil_sample_count, ceil_remaining

    # Iterate over all possible solutions, and track the best solution.
    best_sample_count = None
    best_period_count = None
    best_remaining = None

    n = 1
    while True:
        num_samples, remaining = sampling_n(period, sample_interval, n, max_samples)
        if num_samples > max_samples:
            break

        # For valid sample counts, if there is no best solution, or remaining of the previous
        # solution is worse than remaining of the new solution (over old_n * prev_n periods),
        # select the new solution.
        if num_samples > 0 and (best_sample_count is None or
                abs(best_remaining / best_period_count) > abs(remaining / n)):
            best_sample_count = num_samples
            best_period_count = n
            best_remaining = remaining

        n = n + 1

    if best_sample_count is None:
        return None, None, None # No solution found (max_samples too low)

    approx_freq = best_period_count / (best_sample_count * sample_interval)
    return approx_freq, best_period_count, best_sample_count


def get_function_generator(blocks, config):
    """
    Construct a function generation object.

    @param blocks: Blocks in the specification.
    @type  blocks: C{list} blocks, in particular L{SineFunctionBlock},
                   L{BlockFunctionBlock}, and L{TriangleFunctionBlock}.

    @param config: General configuration of the data acquisition setup.
    @type  config: L{Configuration} block.

    @return: Generator object for generating the functions.
    @rtype:  L{FunctionGenerator}
    """
    sine_funcs = []
    for blk in blocks:
        if isinstance(blk, build_ast.SineFunctionBlock):
            sine_funcs.append(blk)

    generator = FunctionGenerator()
    generator.set_sine_funcs(sine_funcs, config.delay)
    return generator


class FunctionGenerator:
    """
    Class for generating a function generator.

    @ivar sample_arrays: Map of sample array characteristics to index number.
    @ivar sine_array_map: List of sine function with sample array index number to use.

    """
    def __init__(self):
        self.sample_arrays = {}
        self.sine_array_map = []

    def set_sine_funcs(self, sine_funcs, sample_interval):
        """
        Compute characteristics of the sample arrays needed for the sine functions.
        """
        self.sample_arrays = {}
        self.sine_array_map = []

        for sf in sine_funcs:
            result = compute_sine_sampling(1.0 / sf.frequency, sample_interval / 1000000000.0, sf.max_samples)
            if result[0] is None:
                msg = "Cannot generate samples for the sine function, perhaps max sample count is too low?"
                report.report_error([sf.lineno], msg)
                # Program dies

            sample_index = self.sample_arrays.get(result)
            if sample_index is None:
                sample_index = len(self.sample_arrays)
                self.sample_arrays[result] = sample_index

            self.sine_array_map.append((sf, sample_index))

    def generate_sine_samples(self):
        var_defs = codebox.CodeBox()

        arrays = list(sorted((v, k) for k, v in self.sample_arrays.items()))
        for num, (approx_freq, num_periods, sample_count) in arrays:
            if num_periods > 1:
                line = "/* Unscaled sine, {} Hz, {} periods, {} samples */"
            else:
                line = "/* Unscaled sine, {} Hz, {} period, {} samples */"

            var_defs.append(line.format(approx_freq, num_periods, sample_count))
            var_defs.append('static double sine_samples{}[{}];'.format(num, sample_count))

        var_init = codebox.CodeBox()
        var_init.append("static BOOL samples_done = FALSE;")
        var_init.append("if (!samples_done) {", post=1)
        var_init.append("int idx;")
        var_init.append()
        var_init.append("samples_done = TRUE;")
        for num, (_approx_freq, num_periods, sample_count) in arrays:
            var_init.add("for (idx = 0; idx < {}; idx++) {{".format(sample_count), post=1)
            line = 'sine_samples{}[idx] = sin({} * 2 * M_PI * idx / {});'
            var_init.append(line.format(num, num_periods, sample_count))
            var_init.append("}", pre=-1)

        var_init.append("}", pre=-1)

        return var_defs, var_init

    def get_func_block_name(self, shape_name, blk_num):
        return "{}_function{}".format(shape_name, blk_num)

    def generate_block_function(self, obj_prefix, bfunc, blk_num):
        block_var = self.get_func_block_name('block', blk_num)
        var_defs = codebox.CodeBox()
        var_defs.add("BlockFunction {};".format(block_var))

        period = 1000000000 / bfunc.frequency
        hi_length = int(bfunc.duty_cycle * period)
        period = int(period)
        if hi_length < 0:
            hi_length = 0
        elif hi_length > period:
            hi_length = period;

        var_init = codebox.CodeBox()
        line = "InitBlockFunction(&{}{}, {}, {}, {}, {}, 0);"
        var_init.add(line.format(obj_prefix, block_var, hi_length, period, bfunc.high, bfunc.low))

        value = "GetBlockFunctionSample(&{}{}, SOEM_INTERVAL)".format(obj_prefix, block_var)
        return var_defs, var_init, value

    def generate_triangle_function(self, obj_prefix, tfunc, blk_num):
        block_var = self.get_func_block_name('triangle', blk_num)
        var_defs = codebox.CodeBox()
        var_defs.add("TriangleFunction {};".format(block_var))

        period = 1000000000 / tfunc.frequency
        hi_length = int(tfunc.peak_at * period)
        period = int(period)
        if hi_length < 0:
            hi_length = 0
        elif hi_length > period:
            hi_length = period;
        lo_length = period - hi_length

        var_init = codebox.CodeBox()
        line = "InitTriangleFunction(&{}{}, {}, {}, {}, {}, 0);"
        var_init.append(line.format(obj_prefix, block_var, hi_length, lo_length, tfunc.high, tfunc.low))

        value = "GetTriangleFunctionSample(&{}{}, SOEM_INTERVAL)".format(obj_prefix, block_var)
        return var_defs, var_init, value

    def get_sample_count(self, sample_index):
        """
        Find the sample characteristics based on the assigned index number.

        @todo: Horribly slow implementation
        """
        for k, v in self.sample_arrays.items():
            if v == sample_index:
                _approx_freq, _num_periods, sample_count = k
                return sample_count

        assert False, "Unexpected sample index {} found".format(sample_index)

    def get_sample_index(self, sinefunc):
        """
        Get the used sine samples index from the sine function.

        @todo: Horribly slow implementation
        """
        for sf, sample_index in self.sine_array_map:
            if sf == sinefunc:
                return sample_index

        assert False, "Unexpected sine function encountered."


    def generate_sine_function(self, obj_prefix, sfunc, blk_num):
        block_var = self.get_func_block_name('sine', blk_num)
        var_defs = codebox.CodeBox()
        var_defs.add("SineFunction {};".format(block_var))

        sample_index = self.get_sample_index(sfunc)
        sample_count = self.get_sample_count(sample_index)
        line = "InitSineFunction(&{}{}, sine_samples{}, {}, {}, {}, 0);"
        var_init = codebox.CodeBox()
        var_init.add(line.format(obj_prefix, block_var, sample_index, sfunc.base, sfunc.amplitude, sample_count))

        value = "GetSineFunctionSample(&{}{})".format(obj_prefix, block_var)
        return var_defs, var_init, value

