from acqgenlib import codebox, report

def derive_code_filename(code_name, spec_name, ext):
    """
    Compute the name of the output file for writing the generated code to.

    @param code_name: Name of the output file, if available.
    @type  code_name: C{str} or C{None}

    @param spec_name: Name of the input specification.
    @type  spec_name: C{str}

    @param ext: Desired extension to append including leading dot.
    @type  ext: C{str}
    """
    if code_name is not None:
        return code_name

    i = spec_name.rfind(".")
    if i < 0:
        return spec_name + ext
    else:
        return spec_name[:i] + ext

def merge_codelines(*clines):
    """
    Merge code lines maps, named lists of codeboxes.
    """
    result = {}
    for cl in clines:
        for name, lines in cl.items():
            assert name not in result
            result[name] = lines

    return result


def fill_template(codelines, template_fname, lines=False):
    """
    Read the template from file, and insert the codelines into it.

    @param codelines: Mapping of names to lists of code boxes with the content of those names.
    @param template_fname: Name of the template file to read.
    @param lines: If set, generate #line pragmas to show origin of the output.
    """
    skipped = set(codelines.keys())

    code = codebox.CppLineCodeBox(lines)

    try:
        handle = open(template_fname, 'rt')
    except FileNotFoundError as exc:
        msg = "Cannot find template file '{}' (reason: {})".format(template_fname, exc)
        report.report_error(None, msg)

    for templ_lineno, line in enumerate(handle, start=1):
        line = line.rstrip()

        # Copy normal lines from the template..
        if not line.endswith('## */'):
            code.set_fname_line(template_fname, templ_lineno)
            code.abs_add(0, line)
            continue

        pat_line = line.lstrip()
        indent_size = len(line) - len(pat_line)
        # False alarm, it's a normal line after all.
        if not pat_line.startswith('/* ##') or len(pat_line) <= 10:
            code.set_fname_line(template_fname, templ_lineno)
            code.abs_add(0, line)
            continue

        pat_line = pat_line[5:-5].strip()
        clines = codelines.get(pat_line)
        if clines is None:
            print("*** Missing codelines '{}'".format(pat_line))
            code.abs_add(0, line)
            continue

        skipped.discard(pat_line) # Register the codelines with this name was found.

        first = True
        for cl in clines:
            if cl.empty(trim_first=True):
                continue

            if first:
                code.set_fname_line("*template*" + pat_line, 1)
            else:
                code.ensure_empty()
            code.abs_add_box(indent_size, cl)
            first = False

        if first:
            # No content has been added from the codelines, keep original line.
            code.set_fname_line(template_fname, templ_lineno)
            code.abs_add(0, line)

    handle.close()

    for code_name in skipped:
        print("*** Codelines '{}' is not used".format(code_name))

    return code
