INPUT_IOPS  = ['ain', 'din', 'enc']
OUTPUT_IOPS = ['aout', 'dout', 'pwm']

IOPS_WITH_CHANNELS = ['din', 'dout']

class Name:
    """
    Signal name or simple identifier.

    L{iop_name} always exists, in the following cases:
    - If no L{dev_name}, no L{dev_num}, and no L{iop_num} the name represents
      an identifier.
    - If L{dev_num} exists but L{dev_name} does not, it's a generic
      input/output point.
      - If L{iop_num} is also available, it's an io point with channel.
    - If L{dev_name} exists, L{dev_num} also exists, and it's an input/output
      point or channel of a specific device.

    @ivar lineno: Line number of the name in the input specification. Not taken
            into account for equality or hashing.
    @ivar dev_name: Device name, if available.
    @ivar dev_num: Device or io point index number, if available.
    @ivar iop_name: Io point name or identifier.
    @ivar iop_num: Channel number if available.
    """
    def __init__(self, lineno, dev_name, dev_num, iop_name, iop_num):
        self.lineno = lineno
        self.dev_name = dev_name
        self.dev_num  = dev_num
        self.iop_name = iop_name
        self.iop_num  = iop_num

    def classify_name(self):
        """
        Classify the name from its given parts,

        @return: 'device-iop' is a device name (and implied number) are provided.
                 'generic-iop' if no device name, but an io-point name is provided.
                 else 'identifier'.
        """
        if self.dev_name is not None:
            assert self.dev_num is not None
            return "device-iop"
        if self.dev_num is not None:
            return "generic-iop"
        assert self.iop_num is None
        return "identifier"

    def validate(self):
        """
        Validate the name for sanity.

        @return: Error message, if any.
        """
        cl = self.classify_name()
        if cl == 'device-iop':
            # XXX Check device name for being sane. (Not a major problem as
            # device will fail to match at runtime if it is wrong.)

            # Check that the kind has channels if iop_num is set
            if self.iop_num is not None:
                found = False
                for chiop in IOPS_WITH_CHANNELS:
                    if self.iop_name.startswith(chiop):
                        found = True
                        break
                if not found:
                    return "Io kind '{}' has no channels".format(self.iop_name)

            if any(self.iop_name.startswith(kind) for kind in INPUT_IOPS):
                return None # Input kind, channel is always fine.

            # Not an input kind, must be an output kind.
            if any(self.iop_name.startswith(kind) for kind in OUTPUT_IOPS):
                return None

            return "Device io '{}' not recognized".format(self.iop_name)

        if cl == 'generic-iop':
            # Check that the kind has channels if iop_num is set
            if self.iop_num is not None:
                found = False
                for chiop in IOPS_WITH_CHANNELS:
                    if self.iop_name.startswith(chiop):
                        found = True
                        break
                if not found:
                    return "Io kind '{}' has no channels".format(self.iop_name)

            if self.iop_name in INPUT_IOPS:
                return None # Input kind, channel is always fine.

            # Not an input kind, must be an output kind.
            if self.iop_name in OUTPUT_IOPS:
                return None

            return "Generic io '{}' not recognized".format(self.iop_name)

        assert cl == 'identifier'
        if any(self.iop_name.startswith(kind) for kind in INPUT_IOPS):
            return "Missing index of generic input signal"
        if any(self.iop_name.startswith(kind) for kind in OUTPUT_IOPS):
            return "Missing index of generic output signal"
        return None

    def get_inputkind(self):
        """
        Get the kind of input associated with this name.

        @return: Kind of input for this name, or C{None} if not an input.
        @rtype:  C{str} or C{None}
        """
        for iop in INPUT_IOPS:
            if self.iop_name.startswith(iop):
                return iop
        return None

    def get_outputkind(self):
        """
        Get the kind of output associated with this name.

        @return: Kind of output for this name, or C{None} if not an output.
        @rtype:  C{str} or C{None}
        """
        for iop in OUTPUT_IOPS:
            if self.iop_name.startswith(iop):
                return iop
        return None

    def is_partly_write(self):
        if self.iop_num is None:
            return False
        return self.get_outputkind() is not None

    def __str__(self):
        cl = self.classify_name()
        if cl == 'device-iop':
            if self.iop_num is None:
                return "{}[{}]:{}".format(self.dev_name, self.dev_num,
                        self.iop_name)
            else:
                return "{}[{}]:{}.{}".format(self.dev_name, self.dev_num,
                        self.iop_name, self.iop_num)
        if cl == 'generic-iop':
            if self.iop_num is None:
                return "{}[{}]".format(self.iop_name, self.dev_num)
            else:
                return "{}[{}].{}".format(self.iop_name, self.dev_num,
                        self.iop_num)
        assert cl == 'identifier'
        return self.iop_name

    def __hash__(self):
        return (hash(self.dev_name) + hash(self.dev_num) +
                hash(self.iop_name) + hash(self.iop_num))

    def __eq__(self, other):
        if not isinstance(other, Name): return False
        if self.dev_name != other.dev_name: return False
        if self.dev_num != other.dev_num: return False
        if self.iop_name != other.iop_name: return False
        if self.iop_num != other.iop_num: return False
        return True

    def __neq__(self, other):
        return not self.__eq__(other)

def check_names(names):
    """
    Check the provided set of names for correctness and consistency.
    Report errors if problems are detected.

    @precond: L{names} are unique.
    """
    # Check pre-condition
    assert len(set(names)) == len(list(names))

    device_inputkinds = {}
    generic_inputkinds = {}

    device_outputkinds = {}
    generic_outputkinds = {}

    found_error = False
    for name in names:
        # Check the name itself.
        err = name.validate()
        if err is not None:
            msg = "Error: Signal {} at line {}: {}"
            print(msg.format(name, name.lineno, err))
            found_error = True

        kind = name.get_inputkind()
        if kind is not None:
            if name.dev_name is None:
                generic_inputkinds[kind] = name
            else:
                device_inputkinds[kind] = name

            continue

        kind = name.get_outputkind()
        if kind is not None:
            if name.dev_name is None:
                generic_outputkinds[kind] = name
            else:
                device_outputkinds[kind] = name

            continue

        # Not an input or an output -> nothing to check.

    # Check that each iokind is used either as generic or as specific.
    #
    # A generic io-point like "aout[1]" and a specific io-point like
    # "ebox[0].aout1" are considered to be distinct signals in the dependency
    # calculation. Depending on the actual hardware configuration however, this
    # may not be true. To avoid incorrect dependency computations, each input
    # or output kind may only be used in generic or device specific sense.
    #
    # XXX In the future, it may be useful to deal with the possibility of
    # overlapping names.
    #
    def check_common(gen_kinds, dev_kinds):
        gen_keys = set(gen_kinds.keys())
        dev_keys = set(dev_kinds.keys())
        common = gen_keys.intersection(dev_keys)
        if len(common) == 0:
            return True

        element = next(iter(common))
        lines = sorted([gen_kinds[element].lineno, dev_kinds[element].lineno])
        msg = ("Error at lines {} and {}: Using iokind '{}' both in generic " +
               "and device-specific form")
        print(msg.format(lines[0], lines[1], element))
        return False

    if not check_common(generic_inputkinds, device_inputkinds):
        found_error = True
    if not check_common(generic_outputkinds, device_outputkinds):
        found_error = True

    assert not found_error


class Property:
    """
    Parameter property 'name = value' in a block.

    @param lineno: Line number of the property.
    @param name: Name of the property.
    @param type: Kind of the property.
    @param value: Value of the property.
    """
    def __init__(self, lineno, type, name, value):
        self.lineno = lineno
        self.name = name
        self.type = type
        self.value = value
