from acqgenlib.ply import lex, yacc
from acqgenlib import names, report

keywords = {
    'config': 'CONFIG_KW',
    'input': 'INPUT_KW',
    'output': 'OUTPUT_KW',
    'trigger': 'TRIGGER_KW',
    'filter': 'FILTER_KW',
    'function': 'FUNCTION_KW',
}

tokens = list(keywords.values()) + [
    'SQ_OPEN', 'SQ_CLOSE', 'ASG', 'COMMA', 'SEMICOL',
    'PAR_OPEN', 'PAR_CLOSE', 'DOT', 'COLON',
    'IDENTIFIER', 'INTNUMBER', 'REALNUMBER', 'STRING',
]

def t_IDENTIFIER(t):
    '[A-Za-z_][A-Za-z0-9_]*'
    kw = keywords.get(t.value)
    if kw is not None:
        t.type = kw
    return t

def t_STRING(t):
    r'["][^"\n]*["]'
    t.value = t.value[1:-1]
    return t

def t_REALNUMBER1(t):
    r'\d+\.\d+'
    t.value = float(t.value)
    t.type = 'REALNUMBER'
    return t

def t_REALNUMBER2(t):
    r'\d+\.\d+[eE][-+]?\d+'
    t.value = float(t.value)
    t.type = 'REALNUMBER'
    return t

def t_INTNUMBER(t):
    r'\d+'
    t.value = int(t.value)
    return t

def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

t_PAR_OPEN = r'\('
t_PAR_CLOSE = r'\)'
t_SQ_OPEN = r'\['
t_SQ_CLOSE = r'\]'
t_SEMICOL = ';'
t_COLON = ':'
t_COMMA = ','
t_DOT = r'\.'
t_ASG = r'='

t_ignore = ' \t\r'
t_ignore_COMMENT = r'\#.*'

def t_error(t):
    mesg = "Illegal character {}".format(repr(t.value[0]), t.lineno)
    report.report_error([t.lineno], mesg)



def p_program1(p):
    'program :'
    p[0] = []

def p_program2(p):
    'program : program block'
    p[0] = p[1] + [p[2]]

def p_block_output1(p):
    'block : namelist ASG OUTPUT_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('output', p.lineno(2), p[1], p[5][0], p[5][1])

def p_block_output2(p):
    'block : namelist ASG OUTPUT_KW PAR_OPEN PAR_CLOSE SEMICOL'
    p[0] = ('output', p.lineno(2), p[1], [], [])

def p_block_input1(p):
    'block : INPUT_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('input', p.lineno(1), [], p[3][0], p[3][1])

def p_block_function1(p):
    'block : namelist ASG FUNCTION_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('function', p.lineno(2), p[1], p[5][0], p[5][1])

def p_block_trigger1(p):
    'block : namelist ASG TRIGGER_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('trigger', p.lineno(2), p[1], p[5][0], p[5][1])

def p_block_filter1(p):
    'block : namelist ASG FILTER_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('filter', p.lineno(2), p[1], p[5][0], p[5][1])

def p_block_config1(p):
    'block : CONFIG_KW PAR_OPEN parameters PAR_CLOSE SEMICOL'
    p[0] = ('config', p.lineno(1), [], p[3][0], p[3][1])

def p_namelist1(p):
    'namelist : name'
    p[0] = [p[1]]

def p_namelist2(p):
    'namelist : namelist COMMA name'
    p[0] = p[1] + [p[3]]

def p_name1(p):
    'name : IDENTIFIER'
    # Internal signal name, or an identifier.
    p[0] = names.Name(p.lineno(1), None, None, p[1], None)

# Would be a channel of an internal signal name, which is not supported.
#def p_name2(p):
#    'name : IDENTIFIER DOT INTNUMBER'

def p_name3(p):
    'name : IDENTIFIER SQ_OPEN INTNUMBER SQ_CLOSE'
    # Generic input or output, like "dout[0]"
    p[0] = names.Name(p.lineno(1), None, p[3], p[1], None)

def p_name4(p):
    'name : IDENTIFIER SQ_OPEN INTNUMBER SQ_CLOSE DOT INTNUMBER'
    # Generic input with channel, like "din[0].4"
    # Only makes sense for inputs.
    p[0] = names.Name(p.lineno(1), None, p[3], p[1], p[6])

def p_name5(p):
    'name : IDENTIFIER SQ_OPEN INTNUMBER SQ_CLOSE COLON IDENTIFIER'
    # Io point at a specific device "ebox[0]:aout0"
    p[0] = names.Name(p.lineno(1), p[1], p[3], p[6], None)

def p_name6(p):
    'name : IDENTIFIER SQ_OPEN INTNUMBER SQ_CLOSE COLON IDENTIFIER DOT INTNUMBER'
    # Input channel at a specific device "ebox[0]:din0.8"
    # Only makes sense for inputs.
    p[0] = names.Name(p.lineno(1), p[1], p[3], p[6], p[8])

def p_parameters1(p):
    'parameters : nonclosedparameters'
    p[0] = p[1]

def p_parameters2(p):
    'parameters : closedparameters'
    p[0] = p[1]

def p_closedparameters1(p):
    'closedparameters : parameter'
    p[0] = p[1]

def p_closedparameters2(p):
    'closedparameters : nonclosedparameters parameter'
    p[0] = (p[1][0] + p[2][0], p[1][1] + p[2][1])

def p_nonclosedparameters1(p):
    'nonclosedparameters : parameter COMMA'
    p[0] = p[1]

def p_nonclosedparameters2(p):
    'nonclosedparameters : nonclosedparameters parameter COMMA'
    p[0] = (p[1][0] + p[2][0], p[1][1] + p[2][1])

def p_parameter1(p):
    'parameter : name'
    p[0] = ([p[1]], [])

def p_parameter2(p):
    'parameter : IDENTIFIER ASG INTNUMBER'
    p[0] = ([], [names.Property(p.lineno(1), 'intnumber', p[1], p[3])])

def p_parameter3(p):
    'parameter : IDENTIFIER ASG REALNUMBER'
    p[0] = ([], [names.Property(p.lineno(1), 'realnumber', p[1], p[3])])

def p_parameter4(p):
    'parameter : IDENTIFIER ASG name'
    p[0] = ([], [names.Property(p.lineno(1), 'name', p[1], p[3])])

def p_parameter5(p):
    'parameter : IDENTIFIER ASG SQ_OPEN realnumbers SQ_CLOSE'
    p[0] = ([], [names.Property(p.lineno(1), 'reallist', p[1], p[4])])

def p_parameter6(p):
    'parameter : IDENTIFIER ASG STRING'
    p[0] = ([], [names.Property(p.lineno(1), 'string', p[1], p[3])])

def p_realnumbers1(p):
    'realnumbers : REALNUMBER'
    p[0] = [p[1]]

def p_realnumbers2(p):
    'realnumbers : realnumbers COMMA REALNUMBER'
    p[0] = p[1] + [p[3]]


def p_error(p):
    if p:
        lines = [p.lineno]
    else:
        lines = None

    report.report_error(lines, "Syntax error")


def parse_file(fname):
    """
    Parse the configuration file.

    @param fname: Name of the input file.
    @type  fname: C{str}

    @return: Parsed input.
    @rtype:  C{list} of block specifications
    """
    # Read file.
    try:
        handle = open(fname, 'rt', encoding='utf-8')
    except OSError as ex:
        msg = "Cannot open input file '{}' (reason: {})".format(fname, ex.strerror)
        report.report_error(None, msg)

    data = handle.read()
    handle.close()

    # Parse input.
    debug=False
    lexer = lex.lex()
    parser = yacc.yacc(debug=debug, write_tables=False)
    return parser.parse(data, lexer=lexer, debug=debug)

if __name__ == '__main__':
    import sys
    parse_file(sys.argv[1])
