import sys
from acqgenlib import build_ast, report, codebox

def generate_signalgroups(blocks, config):
    """
    Collect and check the signal groups used in the blocks.

    @param blocks: Blocks of the input specification.
    @return: Signal groups ordered by index number.
    @rtype:  C{list} of (C{int}, L{InputBlock}/L{OutputBlock})
    """
    if not config.buffered:
        # For the unbuffered case, don't bother the user with indices, simply
        # assign a number, as it is ignored anyway as there are no ring buffers.
        num = 1
        sorted_groups = []
        for block in blocks:
            if isinstance(block, build_ast.InputBlock) or isinstance(block, build_ast.OutputBlock):
                block.index = num
                sorted_groups.append((num, block))
                num = num + 1

        return sorted_groups


    # Buffered case.
    assert config.buffered

    # Find signal groups, order them on their block number, assuming the same
    # index may be used more than once (use a list for each group number).
    # XXX For efficiency, we should move all input blocks to the front, and all output blocsk to the back.
    # XXX Code doesn't really care about group numbers, but the user does in his matlab file.
    groups = {}
    for block in blocks:
        if isinstance(block, build_ast.InputBlock) or isinstance(block, build_ast.OutputBlock):
            if block.index is None:
                report.report_error([block.lineno], "Missing 'index' number for the group")

            grps = groups.get(block.index)
            if grps is not None:
                grps.append(block)
            else:
                groups[block.index] = [block]

    if len(groups) == 0:
        report.report_error(None, "Missing signal groups")

    found_error = False
    for num, grps in groups.items():
        if len(grps) > 1:
            msg = "Index number {} is used for {} signal groups (only one is allowed)".format(num, len(grps))
            report.report_error([g.lineno for g in grps], msg, keep_running=True)
            found_error = True

    if found_error:
        sys.exit(1)

    # No errors, thus all lists have length 1. Rewrite to a list for sorting.
    sorted_groups = [(num, grps[0]) for num, grps in groups.items()]
    sorted_groups.sort()

    if sorted_groups[0][0] != 1:
        lineno = sorted_groups[0][1].lineno
        msg = "First signal group should have index 1 (found index {})".format(sorted_groups[0][0])
        report.report_error([lineno], msg)

    expected_num = len(sorted_groups)
    number = sorted_groups[-1][0]
    if number != expected_num:
        lineno = sorted_groups[-1][1].lineno
        msg = "Last signal group should have index {} (found index {})".format(expected_num, number)
        report.report_error([lineno], msg)

    return sorted_groups


ML_INPUTGROUP = "MatlabInputGroup{}"
ML_OUTPUTGROUP = "MatlabOutputGroup{}"

# Buffered code generation
# {{{ def generate_ml_input(blknum, grp):
def generate_ml_input(blknum, grp):
    """
    Generate an 'InputGroupN' function that copies a matrix of signal group
    N to a buffer of that signal group.

    @return: Function code as a list of lines in a L{CodeBox}
    """
    rows = len(grp.in_values)
    funcname = ML_INPUTGROUP.format(blknum)

    func_code = codebox.CodeBox()
    func_code.append("/**")
    func_code.append(" * Process signal group {}. Returns \\c true if data has been".format(blknum))
    func_code.append(" * transferred or there is no valid data.")
    func_code.append(" */")
    func_code.append("static BOOL {}(mxArray *outcells)".format(funcname))
    func_code.append("{", post=1)
    func_code.append("Buffer *buffer;")
    func_code.append("mxArray *mat;")
    func_code.append()
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(blknum))
    func_code.append("if (buffer == NULL) {", post=1)
    line = "if (UpdateParentRing(RING{}_PTR, 0) == 0) return FALSE; /* Really no buffer. */"
    func_code.append(line.format(blknum))
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(blknum))
    func_code.append("}", pre=-1)
    func_code.append()
    func_code.append("assert(buffer->used > 0);")
    func_code.append("mat = mxCreateDoubleMatrix({}, buffer->used, mxREAL);".format(rows))
    func_code.append("BufferBlockEmpty(mxGetPr(mat), {}, buffer->used, buffer);".format(rows))
    func_code.append()
    func_code.append("UpdateParentRing(RING{}_PTR, 1);".format(blknum))
    func_code.append()
    func_code.append("mxSetCell(outcells, {}, mat);".format(blknum))
    func_code.append("return TRUE;")
    func_code.append("}", pre=-1)
    return func_code
# }}}
# {{{ def generate_ml_output(blknum, grp):
def generate_ml_output(blknum, grp):
    """
    Generate an 'OutputGroupN' function that copies a buffer of signal group N to a matrix.

    @return: Function code as a list of lines in a L{CodeBox}
    """
    rows = len(grp.out_values)
    funcname = ML_OUTPUTGROUP.format(blknum)

    func_code = codebox.CodeBox()
    func_code.append("/**")
    func_code.append(" * Process signal group {}. Returns \\c true if data has been".format(blknum))
    func_code.append(" * transferred or there is no valid data.")
    func_code.append(" */")
    func_code.append("static BOOL {}(const mxArray *in_mat)".format(funcname))
    func_code.append("{", post=1)
    func_code.append("Buffer *buffer;")
    func_code.append()
    line = "if (in_mat == NULL || !mxIsDouble(in_mat) || mxIsComplex(in_mat) || mxGetM(in_mat) != {}) return TRUE;"
    func_code.append(line.format(rows))
    func_code.append()
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(blknum))
    func_code.append("if (buffer == NULL) {", post=1)
    line = "if (UpdateParentRing(RING{}_PTR, 0) == 0) return FALSE; /* Really no buffer. */"
    func_code.append(line.format(blknum))
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(blknum))
    func_code.append("}", pre=-1)
    func_code.append()
    func_code.append("BufferBlockFill(buffer, mxGetPr(in_mat), {}, mxGetN(in_mat));".format(rows))
    func_code.append("UpdateParentRing(RING{}_PTR, 1);".format(blknum))
    func_code.append("return TRUE;")
    func_code.append("}", pre=-1)
    return func_code
# }}}
# {{{ def generate_ml_io_processing(groups):
def generate_ml_io_processing(groups):
    var_box = codebox.CodeBox()
    body_code = codebox.CodeBox()

    max_block = max(blknum for blknum, _ in groups)
    assert max_block < 64 # bitmask integer size limit.
    if max_block < 32:
        mask_type = "uint32_t"
    else:
        mask_type = "uint64_t"

    # Compute mask constant for the output groups and synchronized input groups.
    allmask, syncmask = 0, 0
    for blknum, grp in groups:
        allmask |= 1 << blknum
        if isinstance(grp, build_ast.OutputBlock):
            syncmask |= 1 << blknum
        elif isinstance(grp, build_ast.InputBlock) and grp.synchronous:
            syncmask |= 1 << blknum

    allmask_text = "0x{:x}".format(allmask)
    syncmask_text = "0x{:x}".format(syncmask)

    # Generate variables.
    var_box.append("ErrorStorage err_store;")
    var_box.append("{} todo;".format(mask_type))
    var_box.append("const mxArray *incells;")
    var_box.append("mxArray *outcells;")
    var_box.append("mxArray *errcells;")
    var_box.append("int index;")
    var_box.append("BOOL result;")

    # Generate body.
    body_code.append("errcells = makeCellArray({});".format(len(groups)))
    body_code.append("outcells = makeCellArray({});".format(len(groups)))
    body_code.append("plhs[0] = errcells;")
    body_code.append("plhs[1] = outcells;")
    body_code.append("incells = prhs[0];")
    body_code.append()

    # Construct text-fragments
    grps_texts = []
    for blknum, grp in groups:
        bitval_text = "0x{:x}".format(1 << blknum)
        todo_cond = "(todo & {}) != 0".format(bitval_text)
        if isinstance(grp, build_ast.InputBlock):
            funcname = ML_INPUTGROUP.format(blknum)
            cells = "outcells"
            iocall_text = "{}({})".format(funcname, cells)
        else:
            funcname = ML_OUTPUTGROUP.format(blknum)
            cells = "incells"
            iocall_text = "{}(mxGetCell({}, {}))".format(funcname, cells, blknum)

        texts = {'todo_cond': todo_cond,
                 'bitval_text': bitval_text,
                 'iocall_text': iocall_text,
                }
        grps_texts.append(texts)

    # Generate looping output drop code.
    body_code.append("todo = {};".format(allmask_text))
    body_code.append("for (;;) { /* Loop until all output is dropped. */", post=1)
    for texts in grps_texts:
        line = "if ({todo_cond} && {iocall_text}) todo ^= {bitval_text};".format_map(texts)
        body_code.append(line)

    body_code.append()
    body_code.append("if ((todo & {}) == 0) break; /* All output and synchronized input done. */".format(syncmask_text))
    body_code.append()
    body_code.append("/* Wait for more space at the output, or for filled buffers at the input. */")
    body_code.append("if (ParentWaitMask(&mesg_box, (todo & {})) != PRS_NOTHING) break;".format(syncmask_text));
    body_code.append("}", pre=-1) # End 'for(;;)'

    body_code.append()
    body_code.append("InitErrorStorage(&err_store);")
    body_code.append("result = ParentPollBox(&mesg_box, &err_store) != PRS_STOP_REALTIME;")
    body_code.append("if (HasMessage(&err_store)) {", post=1)
    body_code.append("index = (err_store.group_index >= 1) ? err_store.group_index : {};".format(max_block + 1))
    body_code.append("mxSetCell(errcells, index, mxCreateString(err_store.message));")
    body_code.append("}", pre=-1)
    body_code.append("return result;");

    return var_box, body_code
# }}}
# {{{ def generate_ml_input_output_funcs(groups):
def generate_ml_input_output_funcs(groups):
    """
    Generate input and output handling function for matlab.
    @return: code for the function foo(incell) -> outcell.
    """

    body_code = codebox.CodeBox()
    for blknum, grp in groups:
        body_code.append()

        if isinstance(grp, build_ast.InputBlock):
            func_code = generate_ml_input(blknum, grp)
        else:
            assert isinstance(grp, build_ast.OutputBlock)
            func_code = generate_ml_output(blknum, grp)

        body_code.add_box(func_code)

    return body_code
# }}}
# {{{ def generate_ringbuffer_code(groups):
def generate_ringbuffer_code(groups):
    num_groups = len(groups)

    ring_vars = codebox.CodeBox()
    ring_funcs = codebox.CodeBox()

    ring_vars.append("static Ringbuffer ringz[{}];".format(num_groups))

    ring_vars.append()
    ring_vars.append("/* Define RINGx macros to compensate for Matlab indexing. */")
    index = 0
    for blknum, _grp in groups:
        ring_vars.append("#define RING{} (ringz[{}])".format(blknum, index))
        ring_vars.append("#define RING{}_PTR (&ringz[{}])".format(blknum, index))
        index = index + 1

    ring_funcs.append("static BOOL InitRings(Memory *mem)")
    ring_funcs.append("{", post=1)
    ring_funcs.append("uint16_t i;")

    size_vars = []
    count_vars = []
    for blknum, grp in groups:
        if isinstance(grp, build_ast.InputBlock):
            name = "input"
            values = grp.in_values
        else:
            name = "output"
            values = grp.out_values

        ring_funcs.append()
        ring_funcs.append("/* {} group {}: {} */".format(name, blknum, ", ".join(str(v) for v in values)))
        ring_funcs.append("uint16_t rows{} = {};".format(blknum, len(values)))
        ring_funcs.append("uint32_t cols{} = {};".format(blknum, grp.buffersize))
        ring_funcs.append("size_t bufsize{} = rows{} * cols{};".format(blknum, blknum,blknum))
        ring_funcs.append("uint16_t bufcount{} = {};".format(blknum, grp.buffercount))
        count_vars.append("bufcount{}".format(blknum))

        ring_funcs.append("size_t size{} = bufsize{} * bufcount{};".format(blknum, blknum, blknum))
        size_vars.append('size{}'.format(blknum))


    ring_funcs.append()
    ring_funcs.append("uint16_t sum_count = {};".format(" + ".join(count_vars)))
    ring_funcs.append("size_t sum_size = {};".format(" + ".join(size_vars)))
    ring_funcs.append("uint16_t buf_num = 0;")
    ring_funcs.append("Buffer *buffers = MemoryAllocate(mem, sizeof(Buffer) * sum_count);")
    ring_funcs.append("double *values = MemoryAllocate(mem, sizeof(double) * sum_size);")
    ring_funcs.append("if (buffers == NULL || values == NULL) return FALSE;")

    for blknum, grp in groups:
        if isinstance(grp, build_ast.InputBlock):
            parent = "FALSE"
        else:
            parent = "TRUE"

        ring_funcs.append()
        ring_funcs.append("for (i = 0; i < bufcount{}; i++) {{".format(blknum), post=1)
        line = "InitBuffer(&buffers[buf_num + i], values, cols{}, rows{});"
        ring_funcs.append(line.format(blknum, blknum))
        ring_funcs.append("values += bufsize{};".format(blknum))
        ring_funcs.append("}", pre=-1)

        line = "InitRingbuffer(RING{}_PTR, &buffers[buf_num], bufcount{}, {});"
        ring_funcs.append(line.format(blknum, blknum, parent))
        ring_funcs.append("buf_num += bufcount{};".format(blknum))

    ring_funcs.append("return TRUE;")
    ring_funcs.append("}", pre=-1)
    ring_funcs.append()

    ring_funcs.append("static void DestroyRings(void)")
    ring_funcs.append("{", post=1)
    for blknum, _grp in groups:
        ring_funcs.append("DestroyRingbuffer(RING{}_PTR);".format(blknum))
    ring_funcs.append("}", pre=-1)
    ring_funcs.append()

    ring_funcs.append("/**")
    ring_funcs.append(" * Test whether one of the rings in the mask has buffers available in the")
    ring_funcs.append(" * parent stage.")
    ring_funcs.append(" * @note: This function may only be used if the tested ring buffers are")
    ring_funcs.append(" *      either not claimed at all, or claimed by the caller.")
    ring_funcs.append(" */")
    ring_funcs.append("BOOL RingsCheckMask(uint64_t mask)")
    ring_funcs.append("{", post=1)
    for blknum, _ in groups:
        count_cond = "(RING{}.outbuffer_count > 0 || RING{}.parent_count > 0)".format(blknum, blknum)
        line = "if ((mask & (1 << {})) != 0 && {}) return TRUE;"
        ring_funcs.append(line.format(blknum, count_cond))

    ring_funcs.append("return FALSE;")
    ring_funcs.append("}", pre=-1)

    return ring_vars, ring_funcs
# }}}

# Unbuffered code generation
def get_unbuf_ingrp_outgrp(groups):
    """
    Get the unbuffered input and the output group.
    """
    in_groups  = [grp for _blknum, grp in groups if isinstance(grp, build_ast.InputBlock)]
    out_groups = [grp for _blknum, grp in groups if isinstance(grp, build_ast.OutputBlock)]
    assert len(in_groups) + len(out_groups) == len(groups)

    if len(in_groups) == 0:
        report.report_error(None, "Missing 'input' block (exactly one input block is required)")
    if len(out_groups) == 0:
        report.report_error(None, "Missing 'output' block (exactly one output block is required)")

    if len(in_groups) > 1:
        linenos = [grp.lineno for grp in in_groups]
        msg = "Expected exactly one 'input' block, found {} blocks"
        report.report_error(linenos, msg.format(len(in_groups)))

    if len(out_groups) > 1:
        linenos = [grp.lineno for grp in out_groups]
        msg = "Expected exactly one 'output' block, found {} blocks"
        report.report_error(linenos, msg.format(len(out_groups)))

    return in_groups[0], out_groups[0]

def generate_unbuf_ml_operational(groups):
    ingrp, outgrp = get_unbuf_ingrp_outgrp(groups)

    ml_op = codebox.CodeBox()
    ml_op.add("/**")
    ml_op.add(" * Perform data IO with SOEM in unbuffered mode.")
    ml_op.add(" * @param plhs Output arguments (always 2).")
    ml_op.add(" * @param in_mat Input matrix with data. Unchecked.")
    ml_op.add(" */")
    ml_op.add("static void MatlabOperational(mxArray *plhs[], const mxArray *in_mat)")
    ml_op.add("{", post=1)
    ml_op.add("const double *inputs;")
    ml_op.add("double *outputs;")
    ml_op.add("const char *err_msg;")
    ml_op.add()

    # Check and setup the 'inputs' value in C for reading the user data.
    in_rows = len(ingrp.in_values)
    ml_op.add('/* Ensure input is a real-valued matrix that is sufficiently large. */')
    line = "if (!mxIsDouble(in_mat) || mxIsComplex(in_mat) || mxGetM(in_mat) < {}) {{"
    ml_op.add(line.format(in_rows), post=1)
    ml_op.add('mexErrMsgIdAndTxt("acq:incorrect:rhs:input:matrix",')
    line = '                  "Input data is not a real valued matrix");'
    ml_op.add(line.format(in_rows))
    ml_op.add("}", pre=-1)

    ml_op.add("inputs = mxGetPr(in_mat);")
    ml_op.add()

    # Setup 'outputs' value in C for writing the results.
    out_rows = len(outgrp.out_values)
    ml_op.add("/* Construct data output matrix. */")
    ml_op.add("plhs[1] = mxCreateDoubleMatrix({}, 1, mxREAL);".format(out_rows))
    ml_op.add('outputs = mxGetPr(plhs[1]);')

    ml_op.add()
    ml_op.add("err_msg = PerformDataCycle(inputs, outputs);")
    ml_op.add("plhs[0] = mxCreateString((err_msg != NULL) ? err_msg : \"\");")
    ml_op.add("}", pre=-1)

    return ml_op

def signalgroups_gencode(groups, config):
    """
    Generate code for the signal groups.

    @param groups: Matlab input and output groups.
    @param config: Configuration.
    @return: Mapping of code-names to lists of code.
    """
    if config.buffered:
        ml_io_funcs = generate_ml_input_output_funcs(groups)
        ml_io_vars, ml_io_code = generate_ml_io_processing(groups)
        ring_vars, ring_funcs = generate_ringbuffer_code(groups)

        return {'ml_input_output_funcs': [ml_io_funcs],
                'ml_io_vars': [ml_io_vars],
                'ml_io_code': [ml_io_code],
                'ringbuffers_variable' : [ring_vars],
                'ringbuffer_funcs': [ring_funcs],
        }

    else:
        # Unbuffered code generation
        ml_op = generate_unbuf_ml_operational(groups)
        return {'ml_operational': [ml_op]}

