/*
 * Generated data acquisition code for buffered input/output.
 */
#include <math.h>
#include "mex.h"
#include "matrix.h"

#include "stdincludes.h"
#include "memory.h"
#include "msgbox.h"
#include "buffer.h"
#include "soem_handling.h"

#include "filterblock.h"
#include "triggerblock.h"
#include "function_generators.h"
#include "nanotime.h"

/** State of the matlab function. */
enum MatlabStateEnum {
    MLS_NOT_INITED,   /**< Before first call to the function. */
    MLS_NOT_REALTIME, /**< Initialized, but no real-time thread running, likely doesn't exist much. */
    MLS_OPERATIONAL,  /**< Function is performing data io. */
};
typedef enum MatlabStateEnum MatlabState;

/* START Global configuration and variables. */
/* ## global_config ## */

static MatlabState ml_state = MLS_NOT_INITED; /**< State of the function. */
static MessageBox mesg_box; /**< ML parent <-> real-time thread message communication. */
static Memory mem_mgr; /**< Memory manager. */
/* END Global configuration and variables. */

/* START Ring buffer variable and functions. */
/* ## ringbuffers_variable ## */
/* ## ringbuffer_funcs ## */
/* END Ring buffer variable and functions. */

/* START Matlab parent state handling. */
static void UnlockMex(void);
static void DestroyDataCycleData(void);

/**
 * Avoid the code from spontaneously disappearing from memory, or at least get
 * a notice beforehand, so the real-time process can be stopped.
 * @return Whether locking succeeded.
 * @pre Data structures should be initialized beforehand.
 */
static BOOL LockMex(void)
{
    void *mem_space;

    assert(ml_state == MLS_NOT_INITED);
    mem_space = mxMalloc(MEMSIZE);
    if (mem_space == NULL) return FALSE;

    if (!mexIsLocked()) {
        mexLock();
        mexAtExit(UnlockMex);
    }

    mexMakeMemoryPersistent(mem_space);
    MemoryInit(&mem_mgr);
    MemorySetMemory(&mem_mgr, mem_space, MEMSIZE);

    MessageBoxInit(&mesg_box);
    InitRings(&mem_mgr);

    return TRUE;

    /* Next, SOEM is started by real-time thread. */
}

/** Cleanup data, uninitialize everything, giving matlab freedom to delete the function. */
static void UnlockMex(void)
{
    if (ml_state == MLS_OPERATIONAL) {
        /* Stop real-time process. */
        ParentStopRealtime(&mesg_box);
        ml_state = MLS_NOT_REALTIME;
    }

    if (ml_state == MLS_NOT_REALTIME) {
        void *mem_space = mem_mgr.base;
        DestroyRings();
        MessageBoxDestroy(&mesg_box);
        MemoryDestroy(&mem_mgr);
        mxFree(mem_space);
        mexUnlock();
    }

    ml_state = MLS_NOT_INITED;
}
/* END Matlab parent state handling. */

/* START Real-time process thread code. */
/* START Io instances of signal groups. */
/** Io instance variables, pointing to the Soem io access points. */
struct IoInstancesStruct {
    /* ## io_variables ## */
};
typedef struct IoInstancesStruct IoInstances;

/**
 * Initialize the #IoInstance variables of #IoInstances, thereby assigning
 * Soem io acces points to signal group variables.
 * @param ssd State and messages of Soem.
 * @param instances Pointer to the io instances.
 * @return Whether initialization was successful.
 */
static const char *InitializeIoInstances(SoemStateData *ssd, IoInstances *instances) {
    /* ## io_var_init ## */
}
static void ClearIoInstances(IoInstances *instances) {
    /* ## clear_io_vars ## */
}
/* END Io instances of signal groups. */
/* START Block variables. */
/** Data owned by the blocks in the specification. */
struct BlockVariablesStruct {
    /* ## block_variables ## */
};
typedef struct BlockVariablesStruct BlockVariables;

/**
 * Initialize the block variables for use.
 * @param blockvars Pointer to the block variables structure.
 * @return Whether initialization was successful.
 */
static BOOL InitializeBlockVariables(BlockVariables *blockvars) {
    /* ## block_init ## */

    return TRUE;
}

static BlockVariables block_vars; /**< Block functions specified by the user. */
static IoInstances io_instances; /** Pointers to IO instances of devices listed by the user. */
static SoemStateData ssd; /**< Soem state. */
static NanoTime increment; /**< Increment step in sample time. */
static NanoTime sample_time; /**< Next sample time. */
static pthread_mutex_t wait_lock; /**<Dummy mutex to block on (for delaying). */
/* END Block variables */

/**
 * Initialize the data cycle data.
 * @return Error message if initialization fails, else \c NULL.
 */
static const char *InitDataCycleData(void)
{
    const NanoTime INITIAL_DELAY = MakeNanos(4, 0); /**< Initial delay 4 seconds. */

    const char *error;

    pthread_mutex_init(&wait_lock, PTHREAD_MUTEX_NORMAL); /* Don't check on recursive locking. */
    pthread_mutex_lock(&wait_lock); /* Make the mutex unavailable. */

    /* Setup soem. */
    InitializeSoemState(&ssd);
    if (!SoemEngineWakeup(&ssd, &mem_mgr, NICS)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }
    if (!SoemSetupIoPoints(&ssd, &mem_mgr)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }

    error = InitializeIoInstances(&ssd, &io_instances);
    if (error != NULL) {
        DestroyDataCycleData();
        return error;
    }
    if (!InitializeBlockVariables(&block_vars)) {
        DestroyDataCycleData();
        return "Failed to initialize block variables";
    }

    if (!SoemEngineStart(&ssd)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }

    sample_time = NowNanos();
    IncreaseNano(&sample_time, &INITIAL_DELAY);
    increment = MakeNano64(SOEM_INTERVAL);

    return NULL;
}

/**
 * Read data values from sensors just retrieved by Soem, and store them into buffers.
 * Retrieve data values for actuator, and insert them into Soem data structures.
 * @param blockvars Structure containing the state of the acquisition blocks.
 * @param instances Pointer to the io instances.
 */
static void RunDataCycle(BlockVariables *blockvars, IoInstances *instances, ErrorStorage *err_out) {
    /* ## block_exec ## */
}

/**
 * Perform data IO cycle.
 * @return \c NULL if communication was ok, else a statically allocated error message.
 */
static void PerformDataCycle(ErrorStorage *err_out)
{
    NanoTime now = NowNanos();

    /* Wait until it is time to sample (code is not expected to actually succeed in locking). */
    pthread_mutex_timedlock(&wait_lock, &sample_time);

    SoemDataIo(&ssd);
    if (ssd.error != NULL) SetError(err_out, ssd.error, -1);
    RunDataCycle(&block_vars, &io_instances, err_out);

    /* Theoretically, this function should be entered before the 'sample_time' deadline.
     * If that didn't happen, sampling is running late by a full sample period or more.
     * Move deadline to entry time to avoid catching up with the past.
     */
    if (CmpNanos(&now, &sample_time) < 0) {
        CopyNanos(&sample_time, &now);
        SetError(err_out, "Sample deadline is moved forward to clock time", -1);
    }

    IncreaseNano(&sample_time, &increment);
}

/** Reset the data used for performing data IO. */
static void DestroyDataCycleData(void)
{
    ClearIoInstances(&io_instances);
    ShutdownSoem(&ssd);

    pthread_mutex_destroy(&wait_lock);
}

/** Real-time thread function. */
static void *RealTimeThread(void *data)
{
    const char *error;
    ErrorStorage err_store;

    InitErrorStorage(&err_store);

    error = InitDataCycleData();
    if (error != NULL) {
        SetError(&err_store, error, -1);
        /* DestroyDataCycleData has already been called if init failed. */
        RealTimeExit(&mesg_box, &err_store);
        return NULL;
    }

    for (;;) {
        PerformDataCycle(&err_store);
        if (RealTimePoll(&mesg_box, &err_store) == PRS_STOP_REALTIME) break;
    }

    DestroyDataCycleData();
    RealTimeExit(&mesg_box, &err_store);
    return NULL;
}
/* END Real-time process thread code. */

/* START Matlab data IO operations. */
/* ## ml_input_output_funcs ## */

/**
 * Construct an 1D cell array.
 * @param length Number of cells of the new array.
 * @return Cell array with the requested number of cells.
 */
static mxArray *makeCellArray(int length)
{
    return mxCreateCellArray(1, &length);
}

static BOOL MatlabOperational(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    /* ## ml_io_vars ## */

    /* Ensure space for 2 outputs exists. */
    if (nlhs != 2) {
        mexErrMsgIdAndTxt("acq:incorrect:lhs:argument:count",
                          "Incorrect number of output arguments, " \
                          "function requires 2 (error and data)");
    }

    /* Ensure input is 1 cell matrix. */
    if (nrhs != 1) {
        mexErrMsgIdAndTxt("acq:incorrect:rhs:argument:count",
                          "Incorrect number of input arguments, " \
                          "function requires 1 cell data argument");
    }
    if (!mxIsCell(prhs[0])) {
        mexErrMsgIdAndTxt("acq:incorrect:rhs:argument:type",
                          "Incorrect type of input argument, " \
                          "function requires 1 cell data argument");
    }

    /* ## ml_io_code ## */
}
/* END Matlab data IO operations. */

/** Standard mex entry point. */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nlhs == 0 && nrhs == 0) { /* Quit action. */
        UnlockMex();
        return;
    }

    if (ml_state == MLS_NOT_INITED) {
        if (!LockMex()) {
            mexErrMsgIdAndTxt("acq:initialization:failed", "Failed to initialize the function");
            return;
        }
        ml_state = MLS_NOT_REALTIME;

        ParentStartRealtime(&mesg_box, &RealTimeThread);
        ml_state = MLS_OPERATIONAL;
    }

    if (!MatlabOperational(nlhs, plhs, nrhs, prhs)) {
        UnlockMex(); /* Real-time closed down, no point in staying up. */
    }
}
