/*
 * Generated data acquisition code for unbuffered input/output.
 */
#include <math.h>
#include "mex.h"
#include "matrix.h"

#include "stdincludes.h"
#include "memory.h"
#include "soem_handling.h"

#include "filterblock.h"
#include "triggerblock.h"
#include "function_generators.h"
#include "nanotime.h"

/** State of the matlab function. */
enum MatlabStateEnum {
    MLS_NOT_INITED,   /**< Before first call to the function. */
    MLS_MEM_LOCKED,   /**< Memory is allocated and locked. */
    MLS_OPERATIONAL,  /**< Function is performing data io. */
};
typedef enum MatlabStateEnum MatlabState;

/* START Global configuration and variables. */
/* ## global_config ## */

static MatlabState ml_state = MLS_NOT_INITED; /**< State of the function. */
static Memory mem_mgr; /**< Memory manager. */
/* END Global configuration and variables. */

/* START Matlab parent state handling. */
static void UnlockMex(void);
static void DestroyDataCycleData(void);

/**
 * Avoid the code from spontaneously disappearing from memory, or at least get
 * a notice beforehand.
 * @return Whether locking succeeded.
 * @pre Data structures should be initialized beforehand.
 */
static BOOL LockMex(void)
{
    void *mem_space;

    assert(ml_state == MLS_NOT_INITED);
    mem_space = mxMalloc(MEMSIZE);
    if (mem_space == NULL) return FALSE;

    if (!mexIsLocked()) {
        mexLock();
        mexAtExit(UnlockMex);
    }

    mexMakeMemoryPersistent(mem_space);
    MemoryInit(&mem_mgr);
    MemorySetMemory(&mem_mgr, mem_space, MEMSIZE);
    ml_state = MLS_MEM_LOCKED;

    return TRUE;

    /* Next, SOEM is started. */
}

/** Cleanup data, uninitialize everything, giving matlab freedom to delete the function. */
static void UnlockMex(void)
{
    if (ml_state == MLS_OPERATIONAL) {
        DestroyDataCycleData();
        ml_state = MLS_MEM_LOCKED;
    }

    if (ml_state == MLS_MEM_LOCKED) {
        void *mem_space = mem_mgr.base;
        MemoryDestroy(&mem_mgr);
        mxFree(mem_space);
        mexUnlock();
    }

    ml_state = MLS_NOT_INITED;
}
/* END Matlab parent state handling. */

/* START Io instances of signal groups. */
/** Io instance variables, pointing to the Soem io access points. */
struct IoInstancesStruct {
    /* ## io_variables ## */
};
typedef struct IoInstancesStruct IoInstances;

/**
 * Initialize the #IoInstance variables of #IoInstances, thereby assigning
 * Soem io acces points to signal group variables.
 * @param ssd State and messages of Soem.
 * @param instances Pointer to the io instances.
 * @return Error message if initialization fails, else \c NULL.
 */
static const char *InitializeIoInstances(SoemStateData *ssd, IoInstances *instances) {
    /* ## io_var_init ## */
}

static void ClearIoInstances(IoInstances *instances) {
    /* ## clear_io_vars ## */
}
/* END Io instances of signal groups. */
/* START Block variables. */
/** Data owned by the blocks in the specification. */
struct BlockVariablesStruct {
    /* ## block_variables ## */
};
typedef struct BlockVariablesStruct BlockVariables;

/**
 * Initialize the block variables for use.
 * @param blockvars Pointer to the block variables structure.
 * @return Whether initialization was successful.
 */
static BOOL InitializeBlockVariables(BlockVariables *blockvars) {
    /* ## block_init ## */

    return TRUE;
}

static BlockVariables block_vars; /**< Block functions specified by the user. */
static IoInstances io_instances; /** Pointers to IO instances of devices listed by the user. */
static SoemStateData ssd; /**< Soem state. */
static NanoTime increment; /**< Increment step in sample time. */
static NanoTime sample_time; /**< Next sample time. */
static pthread_mutex_t wait_lock; /**<Dummy mutex to block on (for delaying). */
/* END Block variables */

/**
 * Read data values from sensors just retrieved by Soem, and store them into buffers.
 * Retrieve data values for actuator, and insert them into Soem data structures.
 * @param blockvars Structure containing the state of the acquisition blocks.
 * @param instances Pointer to the io instances.
 * @param inputs Matrix column supplied by the Matlab application.
 * @param outputs Matrix colmun to return to the Matlab application.
 */
static void RunDataCycle(BlockVariables *blockvars, IoInstances *instances,
                         const double *inputs, double *outputs) {
    /* ## block_exec ## */
}

/** Reset the data used for performing data IO. */
static void DestroyDataCycleData(void)
{
    ClearIoInstances(&io_instances);
    ShutdownSoem(&ssd);

    pthread_mutex_destroy(&wait_lock);
}

/**
 * Initialize the data cycle data.
 * @return Error message if initialization fails, else \c NULL.
 */
static const char *InitDataCycleData(void)
{
    const NanoTime INITIAL_DELAY = MakeNanos(4, 0); /**< Initial delay 4 seconds. */

    const char *error;

    pthread_mutex_init(&wait_lock, PTHREAD_MUTEX_NORMAL); /* Don't check on recursive locking. */
    pthread_mutex_lock(&wait_lock); /* Make the mutex unavailable. */

    /* Setup soem. */
    InitializeSoemState(&ssd);
    if (!SoemEngineWakeup(&ssd, &mem_mgr, NICS)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }
    if (!SoemSetupIoPoints(&ssd, &mem_mgr)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }

    error = InitializeIoInstances(&ssd, &io_instances);
    if (error != NULL) {
        DestroyDataCycleData();
        return error;
    }
    if (!InitializeBlockVariables(&block_vars)) {
        DestroyDataCycleData();
        return "Failed to initialize block variables";
    }

    if (!SoemEngineStart(&ssd)) {
        error = GetSoemStateError(&ssd);
        DestroyDataCycleData();
        return error;
    }

    sample_time = NowNanos();
    IncreaseNano(&sample_time, &INITIAL_DELAY);
    increment = MakeNano64(SOEM_INTERVAL);

    return NULL;
}

/**
 * Perform data IO cycle.
 * @param inputs Matrix column supplied by the Matlab application.
 * @param outputs Matrix colmun to return to the Matlab application.
 * @return \c NULL if communication was ok, else a statically allocated error message.
 */
static const char *PerformDataCycle(const double *inputs, double *outputs)
{
    NanoTime now = NowNanos();
    const char *error;

    /* Wait until it is time to sample (code is not expected to actually succeed in locking). */
    pthread_mutex_timedlock(&wait_lock, &sample_time);

    SoemDataIo(&ssd);
    error = ssd.error;
    RunDataCycle(&block_vars, &io_instances, inputs, outputs);

    /* Theoretically, this function should be entered before the 'sample_time' deadline.
     * If that didn't happen, sampling is running late by a full sample period or more.
     * Move deadline to entry time to avoid catching up with the past.
     */
    if (CmpNanos(&now, &sample_time) < 0) {
        CopyNanos(&sample_time, &now);
        error = "Sampling too late, deadline moved";
    }

    IncreaseNano(&sample_time, &increment);
    return error;
}


/* ## ml_operational ## */

/** Standard mex entry point. */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nlhs == 0 && nrhs == 0) { /* Quit action. */
        UnlockMex();
        return;
    }

    /* Ensure output space exists (error and data). */
    if (nlhs != 2) {
        mexErrMsgIdAndTxt("acq:incorrect:lhs:argument:count",
                          "Incorrect number of output arguments, " \
                          "function requires 2 (error string and real-valued sample data)");
    }

    /* Ensure input exists. */
    if (nrhs != 1) {
        mexErrMsgIdAndTxt("acq:incorrect:rhs:argument:count",
                          "Incorrect number of input arguments, " \
                          "function requires 1 real-valued matrix argument");
    }


    if (ml_state == MLS_NOT_INITED) {
        const char *error;

        if (!LockMex()) {
            mexErrMsgIdAndTxt("acq:initialization:failed", "Failed to initialize the function");
            return;
        }
        /* ml_state has changed to MLS_MEM_LOCKED. */

        error = InitDataCycleData();
        if (error != NULL) {
            UnlockMex();
            /* ml_state has changed to MLS_NOT_INITED. */

            /* Error happened, return error and dummy data array. */
            plhs[0] = mxCreateString(error);
            plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL);
            mxGetPr(plhs[1])[0] = 0.0;
            return;
        }

        ml_state = MLS_OPERATIONAL; /* We're in business */
    }

    MatlabOperational(plhs, prhs[0]);
}
